"""Collection of small functions to simplify common tasks."""

from typing import Callable


def bundle_callables(*callables: Callable) -> Callable:
    """
    Bundle many callables into a single function.

    Callables will be invoked in the order given.
    """

    def bundle_callables_inner(*args, **kwargs):
        """Call all callables in order."""
        for method in callables:
            method(*args, **kwargs)

    return bundle_callables_inner


def bitwise_add(a: int, b: int) -> int:
    """Merge two binary masks together."""
    return a | b


def bitwise_remove(a: int, b: int) -> int:
    """Remove a binary mask from another mask."""
    if bitwise_contains(a, b):
        return a ^ b
    return a


def bitwise_contains(a: int, b: int) -> bool:
    """Return True if any bit of mask `b` is contained in `a`."""
    return bool(a & b)



"""Collection of basic graphical constructs."""

import logging
from abc import abstractmethod
from typing import Any, Tuple

import cocos
from shimmer.display.data_structures import Color
from shimmer.log_utils import log_exceptions

log = logging.getLogger(__name__)


def create_color_rect(width: int, height: int, color: Color) -> cocos.layer.ColorLayer:
    """Create a colored rectangle."""
    log.debug(f"Creating color rect of size ({width=}, {height=}).")
    return cocos.layer.ColorLayer(
        *color.as_tuple_alpha(), width=int(width), height=int(height)
    )


class UpdatingNode(cocos.cocosnode.CocosNode):
    """
    A Node which updates regularly.

    Only performs an update if `self.dirty` is set to True.

    `dirty` can either be set by an third party; or this node will mark itself as dirty
    if the value returned by `self._current_indicator_value()` changes.
    """

    def __init__(self):
        """Create a new UpdatingNode."""
        super(UpdatingNode, self).__init__()
        self.schedule(self.update)
        self.dirty = False

        # Set to globally unique value so first update always happens.
        self.__last_indicator_value: Any = object()

    def _check_if_dirty(self):
        """Mark this node as dirty if the indicator value has changed."""
        indicator_value = self._current_indicator_value()
        if self.__last_indicator_value != indicator_value:
            self.__last_indicator_value = indicator_value
            self.dirty = True

    def _current_indicator_value(self) -> Any:
        """
        Trigger self-updating by changing the value that this returns.

        Implementing this is optional, if not implemented then this object
        will never self-update.
        """
        pass

    @abstractmethod
    def _update(self, dt: float) -> None:
        """Should be overridden to define what this node does in an update."""
        pass

    @log_exceptions(log)
    def update(self, dt: float) -> None:
        """Called on every frame to determine if this node needs to update itself."""
        # Only update if the value has changed to save on drawing performance.
        self._check_if_dirty()
        if self.dirty:
            self.dirty = False
            self._update(dt)


Point2d = Tuple[int, int]



"""An entity is an object that regularly runs its defined actions."""

import asyncio
import logging
import time

from .definition import EntityDefinition

log = logging.getLogger(__name__)


class Entity:
    """
    An object that regularly defines its configured actions.

    Requires an EntityManager to manage it.
    """

    def __init__(self, definition: EntityDefinition) -> None:
        """
        Create a new Entity.

        :param definition: Definition for the entity to use.
        """
        self.definition = definition
        self.running = False
        self._update_interval_s: float = 1
        self._last_update_time_s: float = 0

    @property
    def seconds_since_last_update(self) -> float:
        """
        Number of seconds since this entity was last updated.

        Useful as the update interval is a target and will not be exactly adhered to.
        """
        return time.monotonic() - self._last_update_time_s

    async def _update(self, dt_s: float) -> None:
        """
        Perform the regular actions of this entity.

        :param dt_s: Seconds since the last update was completed.
        """
        pass

    async def run(self) -> None:
        """The loop of this Entity that schedules its updates."""
        self.running = True
        while self.running:
            self._last_update_time_s = time.monotonic()
            if (
                sleep_time_s := self._update_interval_s - self.seconds_since_last_update
                > 0
            ) :
                await asyncio.sleep(sleep_time_s)
            await self._update(self.seconds_since_last_update)



"""
Module defining dialogs.

A dialog is a Window with buttons for a multiple choice question.
For example, an "Are you sure" dialog with yes/no answers.
"""

from dataclasses import replace
from functools import update_wrapper
from typing import Optional, Set

from .button import ButtonDefinition
from .multiple_choice_buttons import (
    MultipleChoiceButtons,
    MultipleChoiceButtonsDefinition,
)
from .question_definition import (
    MultipleChoiceQuestionDefinition,
    MultipleChoiceResponseCallback,
    QuestionCancelledCallback,
)
from .window import WindowDefinition, Window
from ..components.mouse_box import MouseClickEventCallable

AreYouSure = MultipleChoiceQuestionDefinition(
    text="Are you sure?", choices=["Yes", "No"], allow_multiple=False,
)


def callback_with_answer_and_close_window(
    callback: MultipleChoiceResponseCallback, window: Window
) -> MultipleChoiceResponseCallback:
    """Create a callback that closes the window as well as calling the given question callback."""

    def inner(
        currently_selected: Set[str], changed_choice: str, choice_state: bool
    ) -> None:
        callback(currently_selected, changed_choice, choice_state)
        window.kill()

    inner = update_wrapper(inner, callback)
    return inner


def create_multiple_choice_question_dialog(
    definition: MultipleChoiceQuestionDefinition,
    on_answer: MultipleChoiceResponseCallback,
) -> Window:
    """
    Create a window containing a multiple choice question.

    If the question allows multiple answers, a submit button will be shown that closes the window.
    The submit button doe not call any additional callback as each answer choice will have called
    back already.

    If only a single answer is allowed, the window will close as soon as an option is chosen.

    :param definition: Definition of the question.
    :param on_answer: Called when an answer is selected.
    """
    on_close: Optional[MouseClickEventCallable]

    def cancel_on_close() -> MouseClickEventCallable:
        """Wrapper around the mouse event callback on close that ignores all the arguments."""

        def inner(*_, **__):
            if definition.on_cancel is not None:
                definition.on_cancel()

        return inner

    if definition.on_cancel is not None:
        on_close = cancel_on_close()
    else:
        on_close = None

    buttons_defn = MultipleChoiceButtonsDefinition(
        question=definition, button=ButtonDefinition(width=80, height=40),
    )
    buttons = MultipleChoiceButtons(buttons_defn)

    margin_x, margin_y = 15, 15
    window_definition = WindowDefinition(
        title=definition.text,
        width=buttons.rect.width + 2 * margin_x,
        body_height=buttons.rect.height + 2 * margin_y,
        on_close=on_close,
    )
    window = Window(window_definition)

    # Patch the question callback to also close the window when a selection is made.
    buttons.definition = replace(
        buttons.definition,
        question=replace(
            buttons.definition.question,
            on_select=callback_with_answer_and_close_window(on_answer, window),
        ),
    )
    window.add_child_to_body(buttons)

    return window


def create_are_you_sure_dialog(
    on_answer: MultipleChoiceResponseCallback,
    on_cancel: Optional[QuestionCancelledCallback] = None,
) -> Window:
    """
    Create a standard "Are you sure?" dialog.

    Using the `close` button is equivalent to a `No` answer.

    :param on_answer: Method to call when a selection is made.
    :param on_cancel: Method to call if the X button is used instead of answering the question.
    :return: Window containing Yes/No buttons.
    """
    question_defn = replace(AreYouSure, on_cancel=on_cancel)
    return create_multiple_choice_question_dialog(question_defn, on_answer)



"""Manager for many entities. Handles async scheduling of Entities."""

import asyncio
from typing import List, Optional

import janus

from shimmer.engine.entity.entity import Entity


class AsyncManager:
    """
    Async manager for Entitys.

    Provides a sync-async interface to interact with Entities.
    """

    def __init__(self) -> None:
        """Create an AsyncManager."""
        self.entities: List[Entity] = []
        self.running = False
        self.__to_be_started: Optional[janus.Queue[Entity]] = None
        self.__loop: Optional[asyncio.AbstractEventLoop] = None

    def set_event_loop(self, loop: asyncio.BaseEventLoop) -> None:
        """Set the event loop on this manager, and initialise it."""
        self.__loop = loop
        self.__to_be_started = janus.Queue(loop=loop)

    async def run(self):
        """
        Start running all entities managed by this manager.

        Blocks until `self.running` is set to False. This method is intended to be started in a
        thread.
        """
        if self.__loop is None or self.__to_be_started is None:
            raise RuntimeError("AsyncManager has no event loop set!")

        self.running = True
        while self.running:
            new_entity = await self.__to_be_started.async_q.get()
            asyncio.create_task(new_entity.run())
            self.entities.append(new_entity)

    def add_new(self, entity: Entity) -> None:
        """
        Queue a new entity to be added to this manager.

        This is a synchronous interface into the managers async internals.
        """
        if self.__loop is None or self.__to_be_started is None:
            raise RuntimeError("AsyncManager has no event loop set!")

        self.__to_be_started.sync_q.put(entity)



"""Collection of methods for inspecting the GUI state."""

from typing import Type, List, Iterable, Generator, Optional

import cocos
from .components.box import Box


def get_all_nodes_of_type(
    _type: Type[cocos.cocosnode.CocosNode], scene: Optional[cocos.scene.Scene] = None
) -> List[cocos.cocosnode.CocosNode]:
    """
    Discover all nodes in the current cocos Scene that match the given type.

    :param _type: Type to match on, e.g. `Box`.
    :param scene: The cocos Scene to inspect.
        If None, defaults to the current Scene.
    :return: List of matching nodes.
    """

    def matching_type(node: cocos.cocosnode.CocosNode) -> cocos.cocosnode.CocosNode:
        nonlocal _type
        if isinstance(node, _type):
            return node

    if scene is None:
        scene = cocos.director.director.scene

    return scene.walk(matching_type)


def get_boxes_that_intersect_with_rect(
    boxes: Iterable[Box], rect: cocos.rect.Rect
) -> Generator[Box, None, None]:
    """
    Find which of the given boxes intersect with the given rect.

    Intersection is determined in world coordinate space.

    :param boxes: List of boxes to check.
    :param rect: Rect to check intersection with.
    :return: Generator of boxes that intersect with `rect`.
    """
    for test_box in boxes:
        if test_box.world_rect.intersects(rect):
            yield test_box


def get_boxes_that_intersect_with_box(
    boxes: Iterable[Box], box: Box
) -> Generator[Box, None, None]:
    """
    Find which of the given boxes intersect with the given Box.

    Intersection is determined in world coordinate space.

    :param boxes: List of boxes to check.
    :param box: Box to check intersection with.
    :return: Generator of boxes that intersect with `box`.
    """
    yield from get_boxes_that_intersect_with_rect(boxes, box.world_rect)



"""Box that can be dragged, changing the position of it's parent as well."""

import logging
from dataclasses import dataclass, replace
from typing import Optional, Iterable

import cocos
from shimmer.display.alignment import CenterCenter
from shimmer.display.components.box import Box
from shimmer.display.components.mouse_box import (
    MouseBox,
    MouseBoxDefinition,
    EVENT_HANDLED,
)

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class DraggableBoxDefinition(MouseBoxDefinition):
    """
    Definition of a draggable box.

    MouseBox actions will be overwritten when initialising the DraggableBox

    :param snap_boxes: List of Boxes which the draggable anchor will snap to.
        Snapping means when the anchor is dragged over a snap box, their centers will be aligned.
    """

    snap_boxes: Optional[Iterable[Box]] = None


class DraggableBox(MouseBox):
    """Box that can be dragged, changing the position of it's parent as well."""

    def __init__(self, definition: DraggableBoxDefinition):
        """
        Creates a new DraggableBox.

        :param definition: DraggableBoxDefinition defining the shape of the draggable box.
        """
        defn = replace(
            definition,
            on_press=self.start_dragging,
            on_release=self.stop_dragging,
            on_drag=self.handle_drag,
        )
        super(DraggableBox, self).__init__(defn)
        self.definition: DraggableBoxDefinition = self.definition
        self._currently_snapped: bool = False
        self._snap_drag_record: cocos.draw.Vector2 = cocos.draw.Vector2(0, 0)

    def _should_handle_mouse_press(self, buttons: int) -> bool:
        """Should only handle events if this box is attached to something."""
        return self.parent is not None

    def _should_handle_mouse_release(self, buttons: int) -> bool:
        """Should only handle events if this box is attached to something."""
        return self.parent is not None

    def _should_handle_mouse_drag(self) -> bool:
        """Should only handle drag is this box currently is being dragged."""
        return self._currently_dragging

    def start_dragging(
        self, box: "MouseBox", x: int, y: int, buttons: int, modifiers: int,
    ) -> bool:
        """Set this box as being dragged, and reset previous drag knowledge."""
        self._snap_drag_record = cocos.draw.Vector2(0, 0)
        return super(DraggableBox, self).start_dragging(box, x, y, buttons, modifiers)

    def handle_drag(
        self, box: Box, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """While the mouse is pressed on the area, keep updating the position."""
        self._snap_drag_record += (dx, dy)
        self.parent.position += self._snap_drag_record

        if self.definition.snap_boxes is not None:
            # Move the parent by the snap record so we can test if we still intersect a snap point
            # It will get moved back it we still intersect the current snap point.
            # If we don't intersect with any snap point, then we will have just moved the parent
            # by the correct displacement anyway.
            for snap_box in self.definition.snap_boxes:
                if self.world_rect.intersects(snap_box.world_rect):
                    self._align_with_snap_box(snap_box)
                    break
            else:
                # If no longer intersecting any snap boxes, then stop being snapped
                self._snap_drag_record = cocos.draw.Vector2(0, 0)
                self._currently_snapped = False

        return EVENT_HANDLED

    def _align_with_snap_box(self, snap_box: Box) -> None:
        """Align the parent of this DraggableBox with the given snap box."""
        log.debug(f"Aligning {self} with snap box {snap_box}.")
        alignment_required = self.vector_between_anchors(
            snap_box, CenterCenter, CenterCenter
        )
        self.parent.position += alignment_required
        if not self._currently_snapped:
            # Set the drag record to be the inverse motion. This allows the user to drag on/off
            # the snap point repeatedly right at the edge of the snap box.
            self._snap_drag_record = -alignment_required
            self._currently_snapped = True



"""A Box that automatically sets its size to fit the given text."""

import logging
from dataclasses import dataclass, replace
from typing import Optional, Dict, Any

import pyglet
from pyglet import gl

import cocos
from ..alignment import HorizontalAlignment, LeftBottom
from ..components.box import Box, bounding_rect_of_rects
from ..components.focus import KeyboardFocusBox, FocusBoxDefinition, make_focusable
from ..components.font import FontDefinition, Calibri
from ..components.mouse_box import (
    MouseBox,
    MouseBoxDefinition,
    EVENT_UNHANDLED,
    EVENT_HANDLED,
)
from ..keyboard import KeyboardHandlerDefinition, KeyboardHandler

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class TextBoxDefinition(MouseBoxDefinition):
    """
    Definition of a Box containing text. All parameters match pyglet Label parameters.

    Note that the parameters are for the pyglet text layout, rather than the cocos Label.
    The pyglet text layout is a property of the Label called `element.

    Pay attention in particular to `width` and `height`. For this Box they define the size of the
    pyglet text label as well as the surrounding Box. That means that a fixed-size TextBox will
    be very difficult to get exactly centered text. Instead you should consider creating a
    dynamically sized TextBox (i.e. width=None, height=None) and centering that inside another Box.
    """

    text: str = ""
    font: FontDefinition = Calibri

    # Maximum width of the label. Text is wrapped to fit within this width.
    # Set to None to force single line text of arbitrary width.
    width: Optional[int] = None

    # Maximum height of the label. If None, height of the box is set as needed based on
    # the length of text and whether it is multiline or not.
    height: Optional[int] = None

    # Justification the text. E.g. do multiple lines align on the left, center or right?
    justification: HorizontalAlignment = HorizontalAlignment.left

    # Used to override default multiline behaviour which is based on whether `width` is set or not.
    # Use `is_multiline()` to determine whether to use multiline or not.
    multiline: Optional[bool] = None

    def is_multiline(self) -> bool:
        """Determine whether this label should be multiline or not if user hasn't specified."""
        if self.multiline is None:
            return self.width is not None
        return self.multiline

    def to_pyglet_label_kwargs(self) -> Dict[str, Any]:
        """Convert this definition into cocos/pyglet compatible kwargs."""
        return {
            "text": self.text,
            "width": self.width,
            "height": self.height,
            "align": self.justification.value,
            "anchor_x": LeftBottom.x.value,
            "anchor_y": LeftBottom.y.value,
            "multiline": self.is_multiline(),
            **self.font.to_pyglet_label_kwargs(),
        }


class TextBox(MouseBox):
    """
    A rectangle containing text with an optional background color.

    The size of the text box always matches the size of the text (i.e. the size of the
    underlying pyglet Label).

    A TextBox can be thought of as a Box which contains a pyglet Label.
    """

    def __init__(self, definition: TextBoxDefinition):
        """Create a new TextBox."""
        self._label: Optional[cocos.text.Label] = None
        super(TextBox, self).__init__(definition)
        self.definition: TextBoxDefinition = definition
        self._update_label()

    @property
    def text(self) -> str:
        """Get the current text in the Box."""
        return self.definition.text

    @text.setter
    def text(self, value: str) -> None:
        """Set the text of the box and update the display."""
        self.set_text(value)

    def set_text(self, text: str) -> None:
        """Set the text of the box and update the display."""
        self.definition = replace(self.definition, text=text)
        self._update_label()

    def _update_label(self):
        """Update the text of this Box."""
        if self._label is not None:
            self.remove(self._label)

        self._label = cocos.text.Label(**self.definition.to_pyglet_label_kwargs())

        if self.definition.is_dynamic_sized:
            width = self._label.element.content_width
            height = self._label.element.content_height
        else:
            width = self.definition.width
            height = self.definition.height

        # Set the position of the label to the center of this Box.
        # Only attempt to calculate this if the text is not empty, otherwise pyglet gets sad.
        if self.definition.text:
            self._label.position = LeftBottom.get_coord_in_rect(width, height)
        self.add(self._label)

    def bounding_rect_of_children(self) -> cocos.rect.Rect:
        """Get the rect containing all of this boxes children and also the pyglet label."""
        rect = super(TextBox, self).bounding_rect_of_children()
        if self._label is None:
            return rect

        # Pyglet labels are not Boxes, so have to manually calculate the rect of this Box.
        if self.definition.is_dynamic_sized:
            self_rect = cocos.rect.Rect(
                0,
                0,
                self._label.element.content_width,
                self._label.element.content_height,
            )
        else:
            self_rect = self.rect
        return bounding_rect_of_rects((rect, self_rect))


@dataclass(frozen=True)
class EditableTextBoxDefinition(TextBoxDefinition):
    """
    Definition of a user-editable text box.

    If `height` is None, then the height of the font is used.
    """

    width: int = 300
    height: Optional[int] = None
    multiline: bool = False

    @property
    def actual_height(self) -> int:
        """
        Height of the text box to create.

        Equal to `height` if given, otherwise will be the height required for a single line of text.
        """
        return self.height if self.height is not None else self.font.height


class EditableTextBox(MouseBox):
    """
    A text box whose text can be edited by the user.

    The size of this box matches the defined size, unlike TextBox (which matches the
    text size).
    """

    def __init__(self, definition: EditableTextBoxDefinition):
        """Create a new EditableTextBox."""
        definition = replace(
            definition,
            height=definition.actual_height,
            on_press=self.enter_edit_mode,
            on_release=self.exit_drag_mode,
            on_hover=self.set_text_cursor,
            on_unhover=self.set_normal_cursor,
            on_drag=self.highlight_text_on_drag,
        )
        super(EditableTextBox, self).__init__(definition)
        self.definition: EditableTextBoxDefinition = definition
        self.batch = pyglet.graphics.Batch()

        document = pyglet.text.decode_attributed(self.definition.text)
        self._layout = pyglet.text.layout.IncrementalTextLayout(
            document,
            self.definition.width,
            self.definition.actual_height,
            multiline=self.definition.multiline,
            dpi=self.definition.font.dpi,
            batch=self.batch,
        )
        self._caret: pyglet.text.caret.Caret = pyglet.text.caret.Caret(
            self._layout, batch=self.batch
        )
        self._hide_caret()
        self._text_cursor = cocos.director.director.window.get_system_mouse_cursor(
            "text"
        )
        self._keyboard_handler = KeyboardHandler(
            KeyboardHandlerDefinition(
                on_text=self._caret.on_text,
                on_text_motion=self._caret.on_text_motion,
                on_text_motion_select=self._caret.on_text_motion_select,
            )
        )
        self.add(self._keyboard_handler)

        self.focus_box = make_focusable(
            self,
            FocusBoxDefinition(
                on_take_focus=self.take_focus, on_lose_focus=self.lose_focus,
            ),
            focus_type=KeyboardFocusBox,
        )

    def take_focus(self):
        """Start capturing keyboard events."""
        self._show_caret()
        self._keyboard_handler.set_focused()

    def lose_focus(self):
        """Stop capturing keyboard events."""
        self._hide_caret()
        self._keyboard_handler.set_unfocused()

    def draw(self):
        """Draw the pyglet children."""
        gl.glPushMatrix()
        self.transform()
        self.batch.draw()
        gl.glPopMatrix()

    def _show_caret(self):
        """Make the text caret visible."""
        self._caret.visible = True
        self._caret.mark = 0
        self._caret.position = len(self._layout.document.text)

    def _hide_caret(self):
        """Make the text caret invisible."""
        self._caret.visible = False
        self._caret.mark = 0
        self._caret.position = 0

    def enter_edit_mode(
        self, box: "EditableTextBox", x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """Enter edit mode for this EditableTextBox."""
        x, y = self.point_to_local((x, y))
        self.start_dragging(box, x, y, buttons, modifiers)
        self._caret.on_mouse_press(x, y, buttons, modifiers)
        return EVENT_HANDLED

    def exit_drag_mode(
        self, box: "EditableTextBox", x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """Stop dragging - i.e. stop highlighting text."""
        self.stop_dragging(box, x, y, buttons, modifiers)
        return EVENT_HANDLED

    def set_text_cursor(
        self, box: "EditableTextBox", x: int, y: int, dx: int, dy: int
    ) -> Optional[bool]:
        """Make the mouse cursor look like the system text cursor."""
        cocos.director.director.window.set_mouse_cursor(self._text_cursor)
        return EVENT_UNHANDLED

    def set_normal_cursor(
        self, box: "EditableTextBox", x: int, y: int, dx: int, dy: int
    ) -> Optional[bool]:
        """Make the mouse cursor look like the normal system cursor."""
        cocos.director.director.window.set_mouse_cursor(None)
        return EVENT_UNHANDLED

    def highlight_text_on_drag(
        self,
        box: "EditableTextBox",
        x: int,
        y: int,
        dx: int,
        dy: int,
        buttons: int,
        modifiers: int,
    ) -> Optional[bool]:
        """Highlight the text that has been dragged over by the mouse."""
        x, y = self.point_to_local((x, y))
        self._caret.on_mouse_drag(x, y, dx, dy, buttons, modifiers)
        return EVENT_HANDLED


def add_centralised_text(
    box: Box, text: str, text_box_definition: Optional[TextBoxDefinition] = None
) -> None:
    """Add the given text to the center of the given Box."""
    if text_box_definition is None:
        text_box_definition = TextBoxDefinition(text=text)
    else:
        text_box_definition = replace(text_box_definition, text=text)

    text_box = TextBox(text_box_definition)
    box.add(text_box)
    text_box.align_anchor_with_other_anchor(box)



"""
Module defining components that make up a selection system.

This is a system that allows users to define a box using the mouse and get
a list of selectable Boxes inside that.

For example, this is used to make a unit system where the user can select
multiple units at once to give them orders.
"""

import logging
from dataclasses import dataclass, replace
from typing import Optional, Set, cast, Callable

import pyglet

import cocos
from shimmer.display.helpers import bitwise_contains
from .box import Box, BoxDefinition
from .drawing import RectDrawingBoxDefinition, RectDrawingBox, MouseDefinedRect
from ..inspections import get_all_nodes_of_type, get_boxes_that_intersect_with_rect

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class SelectableBoxDefinition(BoxDefinition):
    """
    Definition of a box that is selectable.

    You can specify callbacks for each event type that occurs when a selection is being made.

    Callbacks will receive a MouseDefinedRect detailing the selection that was made, which
    (for example) allows you to determine what action to take based on whicih mouse button
    or keyboard modifiers was used to make the selection.

    If the highlight or selection is set directly without a mouse drawn selection, then
    the callback will receive None instead.
    """

    on_highlight: Optional[Callable[[Optional[MouseDefinedRect]], None]] = None
    on_unhighlight: Optional[Callable[[Optional[MouseDefinedRect]], None]] = None
    on_select: Optional[Callable[[Optional[MouseDefinedRect]], None]] = None
    on_deselect: Optional[Callable[[Optional[MouseDefinedRect]], None]] = None

    # If None, then default behaviour is to call `on_deselect` if no modifier
    # specified in `additive_modifiers` is used.
    # Should return True if the Box remains selected, otherwise False.
    on_new_selection_start: Optional[Callable[[MouseDefinedRect], bool]] = None

    # Bitwise sum of the pyglet representations of modifier keys (e.g. SHIFT) that cause
    # additional selection drawings to add to the currently selected set.
    # If the modifier isn't used, then when a new selection is started, the previous
    # selection will be deselected.
    # If multiple modifiers are allowed, then any one of them can be used to trigger
    # additive behaviour (i.e. not all are required).
    # Set to 0 to exclude all modifiers.
    additive_modifiers: int = pyglet.window.key.MOD_SHIFT


class SelectableBox(Box):
    """
    A Box that can be selected.

    Typical usage is to add this Box to another Box that you want to be selectable
    using the mouse. Define callbacks in the definition to control what happens when this
    Box is selected or highlighted.

    See SelectionDrawingBox to define the area in which boxes can be selected using the mouse.
    """

    def __init__(self, definition: SelectableBoxDefinition):
        """
        Create a new SelectableBox.

        :param definition: Definition of the actions to take when this Box is interacted with.
        """
        super(SelectableBox, self).__init__(definition)
        self.definition: SelectableBoxDefinition = definition
        self._selected: bool = False
        self._highlighted: bool = False

    @property
    def selected(self) -> bool:
        """True if this box is currently Selected."""
        return self._selected

    @selected.setter
    def selected(self, value: bool) -> None:
        """
        Set whether this box is selected or not.

        Calls `on_select` or `on_deselect` callbacks if value is changed.
        """
        if value and not self._selected:
            self.on_select()
        elif not value and self._selected:
            self.on_deselect()

    @property
    def highlighted(self) -> bool:
        """True if this box is currently highlighted."""
        return self._highlighted

    @highlighted.setter
    def highlighted(self, value: bool) -> None:
        """
        Set whether this box is highlighted or not.

        Calls `on_highlight` or `on_unhighlight` callbacks if value is changed.
        """
        if value and not self._highlighted:
            self.on_highlight()
        elif not value and self._highlighted:
            self.on_unhighlight()

    def on_highlight(self, selection_rect: Optional[MouseDefinedRect] = None) -> None:
        """
        Called when a selection is not complete, but includes this Box.

        For example, this may be used when a user is definition a selection by
        dragging the mouse, and that selection is drawn overlapping this box's rect.
        """
        log.debug(f"{self} has been selection highlighted.")
        self._highlighted = True
        if self.definition.on_highlight is not None:
            self.definition.on_highlight(selection_rect)

    def on_unhighlight(self, selection_rect: Optional[MouseDefinedRect] = None) -> None:
        """Called when a selection is not complete, and no longer includes this Box."""
        log.debug(f"{self} has been selection unhighlighted.")
        self._highlighted = False
        if self.definition.on_unhighlight is not None:
            self.definition.on_unhighlight(selection_rect)

    def on_select(self, selection_rect: Optional[MouseDefinedRect] = None) -> None:
        """
        Called when a selection including this Box is completed.

        It is not guaranteed that `on_unhighlight` will be called before `on_select` is called.
        """
        log.debug(f"{self} has been selected.")
        self._selected = True
        if self.definition.on_select is not None:
            self.definition.on_select(selection_rect)

    def on_deselect(self, selection_rect: Optional[MouseDefinedRect] = None) -> None:
        """
        Called when the selection of this Box is removed.

        :param selection_rect: Rect of the selection event that occurred to trigger deselection.
            Typically will not intersect with this Box.
        """
        log.debug(f"{self} has been deselected.")
        self._selected = False
        if self.definition.on_deselect is not None:
            self.definition.on_deselect(selection_rect)

    def on_new_selection_start(self, selection_rect: MouseDefinedRect) -> bool:
        """
        Called when a new selection box starts being drawn if this Box is already selected.

        Default behaviour is to call on_deselect if the modifier key is not being used.

        :param selection_rect: The selection rect that is just being started. Typically will
            not intersect with this Box.
        :return: True if this box remains selected, otherwise False.
        """
        if self.definition.on_new_selection_start is None:
            if not bitwise_contains(
                selection_rect.modifiers, self.definition.additive_modifiers
            ):
                self.on_deselect(selection_rect)
                return False
            return True
        return self.definition.on_new_selection_start(selection_rect)


class SelectionDrawingBox(RectDrawingBox):
    """
    A box that within its boundaries, any SelectableBox can be selected.

    This allows the user to drag to highlight or select many boxes at once.
    For example, this allows creation of a unit selection system.
    """

    def __init__(
        self,
        definition: Optional[RectDrawingBoxDefinition] = None,
        rect: Optional[cocos.rect.Rect] = None,
    ):
        """
        Create a new SelectionDrawingBox.

        :param definition: Definition of this drawing box. The callbacks will be overridden.
        :param rect: Definition of the rectangle that a drag selection can be made within.
            If None, defaults to the entire window.
        """
        if definition is None:
            definition = RectDrawingBoxDefinition()

        definition = replace(
            definition,
            on_start=self.cache_selectable_boxes,
            on_change=self.handle_incomplete_selection_change,
            on_complete=self.handle_complete_selection,
        )
        super(SelectionDrawingBox, self).__init__(definition, rect)
        self._cache: Set[SelectableBox] = set()
        self._pending_selection: Set[SelectableBox] = set()
        self._current_selection: Set[SelectableBox] = set()

    def cache_selectable_boxes(self, defined_rect: MouseDefinedRect) -> None:
        """
        Cache the possible selectable boxes to speed up selection detection.

        This stores a reference to all SelectableBoxes in the current scene, so we don't
        have to re-find them on every selection box change.

        This does mean that new SelectableBoxes created while a selection is being made will
        be missed, but that is unlikely - so prefer the faster speed of using a cache.
        """
        self._cache = set(get_all_nodes_of_type(SelectableBox))
        self._pending_selection = set()

        # Tell each box in the last completed selection that we are starting a new selection.
        # Those boxes might want to deselect themselves.
        # Remove those which do deselect themselves from the current selection, and keep the rest.
        to_remove = list()
        for box in self._current_selection:
            still_selected = box.on_new_selection_start(defined_rect)
            if not still_selected:
                to_remove.append(box)

        for box in to_remove:
            self._current_selection.remove(box)

    def handle_incomplete_selection_change(
        self, defined_rect: MouseDefinedRect
    ) -> None:
        """
        Called when the pending selection changes (i.e. mouse dragged).

        Call on_highlight on each SelectableBox that is now in the selection area,
        and on_unhighlight on each SelectableBox that is no longer in the selection area.
        """
        intersected_boxes = set(
            get_boxes_that_intersect_with_rect(
                self._cache, defined_rect.as_world_rect()
            )
        )
        intersected_boxes = cast(Set[SelectableBox], intersected_boxes)

        # For every box in the intersection, but not in the current selection, highlight it.
        for box in intersected_boxes.difference(self._pending_selection):
            self._pending_selection.add(box)
            if not box.selected:
                box.on_highlight(defined_rect)

        # For every box in the current selection, but not in the intersection, unhighlight it.
        for box in self._pending_selection.difference(intersected_boxes):
            self._pending_selection.remove(box)
            if not box.selected:
                box.on_unhighlight(defined_rect)

    def handle_complete_selection(self, defined_rect: MouseDefinedRect) -> None:
        """
        Called when the selection is completed (i.e. mouse released).

        Call on_select on each SelectableBox that is now in the selection area,
        and on_unhighlight on each SelectableBox that is no longer in the selection area.
        """
        intersected_boxes = set(
            get_boxes_that_intersect_with_rect(
                self._cache, defined_rect.as_world_rect()
            )
        )
        intersected_boxes = cast(Set[SelectableBox], intersected_boxes)

        self._current_selection.update(intersected_boxes)

        # For every box in the pending selection, but not in the final selection, unhighlight it.
        for box in intersected_boxes.difference(self._pending_selection):
            if box.highlighted and not box.selected:
                box.on_unhighlight(defined_rect)
        self._pending_selection = set()

        # Tell each Box that is has been selected.
        for box in self._current_selection:
            if not box.selected:
                box.on_select(defined_rect)



"""Widgets for asking questions of the user."""

from dataclasses import dataclass, field
from typing import Sequence, Optional, Protocol, Set, Union

from ..components.box import Box


class QuestionCancelledCallback(Protocol):
    """Protocol defining the signature of a callback when a question is cancelled."""

    def __call__(self) -> None:
        """Function signature of a question cancelled callback."""
        pass


class MultipleChoiceResponseCallback(Protocol):
    """Protocol defining the signature of a callback when a multiple choice option is selected."""

    def __call__(
        self,
        currently_selected: Set[Union[str, Box]],
        changed_choice: Union[str, Box],
        choice_state: bool,
    ) -> Optional[bool]:
        """Function signature of a multiple choice response callback."""
        pass


@dataclass(frozen=True)
class QuestionDefinition:
    """Definition of a question requesting user input."""

    text: str

    # Callback that is called if a question is cancelled.
    # For example, this could be when a Window displaying a question is closed using the X button.
    on_cancel: Optional[QuestionCancelledCallback] = None


@dataclass(frozen=True)
class MultipleChoiceQuestionDefinition(QuestionDefinition):
    """
    Definition of a question with a defined set of answers.

    Users may select multiple answers if `allow_multiple` is set to True.
    """

    # The choices to include. Ordering impacts the way the question answer will be presented to the
    # user.
    choices: Sequence[Union[str, Box]] = field(default_factory=list)

    # Whether to allow multiple options to be selected at once.
    allow_multiple: bool = False

    # Set of choices that will be selected on creation.
    defaults: Set[Union[str, Box]] = field(default_factory=set)

    # Callback to call when the selection changes.
    # Arguments are:
    #  - List of the choices that are currently selected.
    #  - The choice that has changed state.
    #  - The current state of that choice.
    on_select: Optional[MultipleChoiceResponseCallback] = None

    def __post_init__(self):
        """Validate the definition."""
        if not self.allow_multiple:
            if len(self.defaults) > 1:
                raise ValueError(
                    "Cannot have multiple defaults if `allow_multiple` is False."
                )
        if not self.defaults.issubset(set(self.choices)):
            invalid_defaults = self.defaults.difference(set(self.choices))
            raise ValueError(
                f"All defaults must be in choices. Invalid defaults: {invalid_defaults}"
            )



"""Definition of code blocks."""

from dataclasses import dataclass, field
from textwrap import indent
from typing import List, Callable, Any, Optional, Tuple


NOT_YET_RUN = object()


def return_true() -> bool:
    """Return True."""
    return True


@dataclass
class Instruction:
    """A single code instruction."""

    method: Callable
    args: Tuple = field(default_factory=tuple)
    result: Optional[Any] = NOT_YET_RUN

    on_execute_start: Optional[Callable] = None
    on_execute_complete: Optional[Callable] = None

    def execute(self):
        """Run the instruction."""
        if self.on_execute_start is not None:
            self.on_execute_start()
        self.result = self.method(*self.args)
        if self.on_execute_complete is not None:
            self.on_execute_complete()

    def method_str(self) -> str:
        """Return the string representation of the instruction."""
        result_str = f" -> {self.result}" if self.result is not NOT_YET_RUN else ""
        arg_str = ", ".join((str(arg) for arg in self.args))
        return f"{self.method.__name__}({arg_str}){result_str}"

    def __str__(self) -> str:
        """Return the string representation of the instruction."""
        return self.method_str()


@dataclass
class CodeBlock:
    """A collection of instructions."""

    instructions: List[Instruction] = field(default_factory=list)

    def run(self) -> None:
        """Run all the instructions sequentially."""
        for instruction in self.instructions:
            instruction.execute()

    def __str__(self) -> str:
        """Return the string representation of all instructions."""
        return "\n".join((str(ins) for ins in self.instructions))


@dataclass
class InstructionWithCodeBlock(Instruction):
    """
    Definition of an instruction with a code block.

    Generally won't be used directly, but is a common base class for If/While instructions.
    """

    code_block: CodeBlock = field(default_factory=CodeBlock)


@dataclass
class If(InstructionWithCodeBlock):
    """
    Definition of an If block.

    Built of:
      - A single boolean instruction.
      - A code block to execute if that instruction returns True.
    """

    def execute(self):
        """Run the If statement. Only run the code_block if the statement returns True."""
        super(If, self).execute()
        if self.result:
            if self.on_execute_start is not None:
                self.on_execute_start()
            self.code_block.run()
            if self.on_execute_complete is not None:
                self.on_execute_complete()

    def method_str(self) -> str:
        """Return the string representation of the If statement without the code block."""
        return "if {}:".format(super(If, self).method_str())

    def __str__(self) -> str:
        """Return the string representation of the If statement and code block."""
        instruction_str = super(If, self).__str__()
        code_block_str = indent(str(self.code_block), "    ")
        return f"{instruction_str}\n{code_block_str}"


@dataclass
class Else(If):
    """An Else block. Only for use with an Elif block."""

    # We're treating an Else block like an If block that always return True.
    method: Callable = return_true

    def method_str(self) -> str:
        """Return the string representation of the header of the else block."""
        return "else:"


@dataclass
class Elif(If):
    """
    An Elif code block.

    Not intended to be used directly, should be used as part of an IfElifElse.
    """

    def method_str(self) -> str:
        """Return the string representation of the Elif statement without the code block."""
        if_str = super(Elif, self).method_str()
        return "el{}".format(if_str)


@dataclass
class IfElifElse(If):
    """
    An If/Elif/Else block.

    Optionally has 0 or many Elif blocks.
    Optionally has 0 or 1 Else block.
    """

    elifs: List[Elif] = field(default_factory=list)

    # Default to a blank Else block which performs no action.
    else_: Else = field(default_factory=Else)

    def execute(self):
        """Run the Elif block."""
        super(IfElifElse, self).execute()
        if not self.result:  # Result of super If statement
            for if_ in self.elifs:
                if_.execute()
                if if_.result:
                    break
            else:  # If no if_ returned True, then call the else.
                self.else_.execute()

    def __str__(self) -> str:
        """Return the string representation of the Elif block."""
        result = super(IfElifElse, self).__str__() + "\n"
        for _elif in self.elifs:
            result += str(_elif) + "\n"
        result += str(self.else_)
        return result


@dataclass
class Program:
    """
    Definition of a runnable program.

    Constructed of a code block and attributes naming this program.
    """

    name: str
    code_block: CodeBlock

    def run(self) -> None:
        """Run the programs code block."""
        self.code_block.run()



"""
A visible, clickable button.

Changes color when interacted with (clicked. on hover) and calls the defined methods.
"""

import logging
from dataclasses import dataclass
from typing import Optional

import cocos
from shimmer.display.components.mouse_box import (
    MouseBoxDefinition,
    MouseBox,
    EVENT_HANDLED,
)
from shimmer.display.data_structures import Color, PassiveBlue, ActiveBlue, MutedBlue
from shimmer.display.helpers import bitwise_contains
from shimmer.display.keyboard import (
    KeyboardActionDefinition,
    KeyboardHandlerDefinition,
    KeyboardHandler,
)
from shimmer.display.widgets.text_box import TextBoxDefinition, TextBox

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class ButtonDefinition(MouseBoxDefinition):
    """A Button definition with visual elements."""

    text: Optional[str] = None

    base_color: Color = PassiveBlue
    depressed_color: Optional[Color] = MutedBlue
    hover_color: Optional[Color] = ActiveBlue

    keyboard_shortcut: Optional[str] = None

    @property
    def formatted_text(self) -> Optional[str]:
        """
        Get the formatted text for this Button.

        When used in a RichLabel, underlines the keyboard shortcut if present in the text.
        """
        if (
            self.text is not None
            and self.keyboard_shortcut is not None
            and self.keyboard_shortcut in self.text
            and len(self.text) > 1
        ):
            # If the keyboard_shortcut appears in the text, and it's not the only character
            # then underline it.
            # This uses pyglets special in-built text formatting engine.
            # https://pyglet.readthedocs.io/en/stable/programming_guide/text.html#attributed-text
            index = self.text.index(self.keyboard_shortcut)
            underline_color = (255, 255, 255, 255)
            return "".join(
                (
                    self.text[:index],
                    f"{{underline {underline_color}}}"
                    f"{self.keyboard_shortcut}"
                    f"{{underline False}}",
                    self.text[index + 1 :],
                )
            )
        return self.text


class Button(MouseBox):
    """
    A Button that has visual elements.

    Changes color when hovered over or clicked on.
    """

    def __init__(self, definition: ButtonDefinition):
        """
        Create a new Button.

        :param definition: Definition of the button.
        """
        super(Button, self).__init__(definition)
        self.definition: ButtonDefinition = self.definition
        self.label: Optional[TextBox] = None
        self.keyboard_handler: Optional[KeyboardHandler] = None
        self.update_label()
        self._set_background_color(self.definition.base_color)
        self.update_keyboard_handler()

    @property
    def rect(self) -> cocos.rect.Rect:
        """Get the rect defining the shape of this button."""
        return self._rect

    def update_label(self):
        """Recreate the button label."""
        if self.label is not None:
            # Don't resize because the label is what controls a dynamically sized button,
            # so save the resizing to when we re-make the label.
            # If the label is just totally removed then there isn't a sensible size to resize
            # to anyway.
            self.remove(self.label, no_resize=True)

        if self.definition.formatted_text is None:
            self.label = None
            return

        self.label = TextBox(TextBoxDefinition(text=self.definition.formatted_text))
        self.add(self.label)
        if not self.definition.is_dynamic_sized:
            # If dynamic sized, then the label helps control setting the size, so it makes no
            # sense to align it.
            self.label.align_anchor_with_other_anchor(self)

    def _should_handle_mouse_press(self, buttons: int) -> bool:
        """Whether this button should handle mouse press events."""
        return (
            self.definition.on_press is not None
            # Also handle if on_release is defined so we can record which mouse button was used.
            or self.definition.on_release is not None
            or self.definition.depressed_color is not None
        )

    def _should_handle_mouse_release(self, buttons: int) -> bool:
        """Whether this button should handle mouse release events."""
        return (
            # Also handle if on_press is defined so we can record which mouse button was used.
            self.definition.on_press is not None
            or self.definition.on_release is not None
            or self.definition.depressed_color is not None
        ) and bitwise_contains(self._currently_pressed, buttons)

    def _should_handle_mouse_hover(self) -> bool:
        """Whether this button should handle mouse hover and unhover events."""
        return (
            self.definition.on_hover is not None
            or self.definition.on_unhover is not None
            or self.definition.hover_color is not None
        )

    def _on_press(self, x: int, y: int, buttons: int, modifiers: int) -> Optional[bool]:
        """Change the button color and call the on_select callback."""
        super(Button, self)._on_press(x, y, buttons, modifiers)
        if self.definition.depressed_color is not None:
            self._set_background_color(self.definition.depressed_color)
        return EVENT_HANDLED

    def _on_release(
        self, x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """Change the button color and call the on_release callback."""
        super(Button, self)._on_release(x, y, buttons, modifiers)
        if self._currently_hovered:
            if self.definition.hover_color is not None:
                self._set_background_color(self.definition.hover_color)
        else:
            self._set_background_color(self.definition.base_color)
        return EVENT_HANDLED

    def _on_hover(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """Change the button color and call the on_hover callback."""
        result = super(Button, self)._on_hover(x, y, dx, dy)
        if self.definition.hover_color is not None:
            self._set_background_color(self.definition.hover_color)
        return result

    def _on_unhover(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """Change the button color and call the on_unhover callback."""
        result = super(Button, self)._on_unhover(x, y, dx, dy)
        if self.definition.hover_color is not None:
            self._set_background_color(self.definition.base_color)
        return result

    def on_keyboard_hover(self) -> Optional[bool]:
        """Called when the keyboard is used to navigate onto this button."""
        return self._on_hover(0, 0, 0, 0)

    def on_keyboard_unhover(self) -> Optional[bool]:
        """Called when the keyboard is used to navigate off this button."""
        return self._on_unhover(0, 0, 0, 0)

    def on_keyboard_select_press(self) -> Optional[bool]:
        """Called when the keyboard is used to press this button."""
        return self._on_press(0, 0, 0, 0)

    def on_keyboard_select_release(self) -> Optional[bool]:
        """Called when the keyboard is used to release this button."""
        return self._on_release(0, 0, 0, 0)

    def update_keyboard_handler(self):
        """
        Recreate the keyboard handler for this Button.

        Listens for keyboard presses for the configured keyboard_shortcut and
        activates on_select and on_release for this button on each key press.
        """
        self.keyboard_handler = None

        if self.definition.keyboard_shortcut is None:
            return

        keymap = KeyboardHandlerDefinition()
        keyboard_action = KeyboardActionDefinition(
            chords=[
                self.definition.keyboard_shortcut.upper(),
                self.definition.keyboard_shortcut.lower(),
            ],
            on_press=self.on_keyboard_select_press,
            on_release=self.on_keyboard_select_release,
        )
        keymap.add_keyboard_action(keyboard_action)
        self.keyboard_handler = KeyboardHandler(keymap)
        self.add(self.keyboard_handler)


class ToggleButton(Button):
    """A button that toggles on and off."""

    def __init__(self, definition: ButtonDefinition):
        """
        Create a new ToggleButton.

        :param definition: Definition of the button.
        """
        super(ToggleButton, self).__init__(definition)
        self._is_toggled: bool = False

    @property
    def is_toggled(self) -> bool:
        """Whether this button is currently toggled on."""
        return self._is_toggled

    @is_toggled.setter
    def is_toggled(self, value: bool) -> None:
        """
        Set the toggled state of this button.

        Calls the on_select / on_release as appropriate.
        """
        if self._is_toggled != value:
            # If requested state is different to current state, then toggle this button.
            self._on_press(0, 0, 0, 0)
        # Otherwise no action needed to toggle the button.

    def _on_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        """
        Called when the Box is clicked by the user.

        Alternatively calls the `on_select` and `on_release` callback from the definition on each
        press.
        """
        self._is_toggled = not self._is_toggled
        if not self._is_toggled:
            super(ToggleButton, self)._on_release(x, y, buttons, modifiers)
        else:
            super(ToggleButton, self)._on_press(x, y, buttons, modifiers)

    def _should_handle_mouse_release(self, buttons: int) -> bool:
        """
        Toggle buttons do no react to mouse release events.

        Super of `on_release` is called when button is pressed while already toggled.
        """
        return False

    def _on_unhover(self, x: int, y: int, dx: int, dy: int) -> None:
        """Change the button color and call the on_unhover callback."""
        super(Button, self)._on_unhover(x, y, dx, dy)
        if self.definition.hover_color is not None:
            # Reset the color back to what is should be based on the toggled state.
            if not self._is_toggled:
                self._set_background_color(self.definition.base_color)
            elif self.definition.depressed_color is not None:
                self._set_background_color(self.definition.depressed_color)



"""Definition of a Box that responds in a user-defined way to mouse events."""

import logging
from dataclasses import dataclass, field
from typing import Optional, Protocol

from pyglet.event import EVENT_UNHANDLED, EVENT_HANDLED

import cocos
from shimmer.display.components.box import ActiveBox, BoxDefinition
from shimmer.display.helpers import bitwise_add, bitwise_remove, bitwise_contains
from shimmer.display.primitives import Point2d
from shimmer.log_utils import LTRACE

log = logging.getLogger(__name__)


class MouseClickEventCallable(Protocol):
    """Protocol defining the signature of on_select and on_release callbacks."""

    def __call__(
        self,
        box: "MouseBox",  # This is the MouseBox that handled the event.
        x: int,
        y: int,
        buttons: int,
        modifiers: int,
    ) -> Optional[bool]:
        """
        The signature of mouse click event callbacks.

        Return True or None to consume the mouse event.
        Return False to allow it to propagate to other handlers.
        """


class MouseMotionEventCallable(Protocol):
    """Protocol defining the signature of on_hover and on_unhover callbacks."""

    def __call__(
        self,
        box: "MouseBox",  # This is the MouseBox that handled the event.
        x: int,
        y: int,
        dx: int,
        dy: int,
    ) -> Optional[bool]:
        """
        The signature of mouse motion event callbacks.

        Important note: The default (i.e. return None) is that mouse motion callbacks do not
            consume the event, which is different to Click or Drag events.

        Return True to consume the mouse event.
        Return False or None to allow it to propagate to other handlers.
        """


class MouseDragEventCallable(Protocol):
    """Protocol defining the signature of on_drag callback."""

    def __call__(
        self,
        box: "MouseBox",  # This is the MouseBox that handled the event.
        x: int,
        y: int,
        dx: int,
        dy: int,
        buttons: int,
        modifiers: int,
    ) -> Optional[bool]:
        """
        The signature of mouse drag event callbacks.

        Return True or None to consume the mouse event.
        Return False to allow it to propagate to other handlers.
        """


@dataclass(frozen=True)
class MouseBoxDefinition(BoxDefinition):
    """
    Definition of an invisible Box that can be clicked and hovered over.

    Method definitions are optional and are called with the mouse event arguments as keywords.

    :param on_press: Called when a mouse button is pressed within the Box.
    :param on_press_outside: Called when a mouse button is pressed outside of the Box.
        Cannot cause mouse events to be consumed.
    :param on_release: Called when a mouse button is released within the Box.
        Note: This may not always follow an `on_select` because the user may drag outside the box
        before releasing the mouse button.
    :param on_hover: Called when the mouse enters the Box.
    :param on_unhover: Called when the mouse leaves the Box.
    :param on_motion: Called when the mouse moves within the Box.
    :param on_drag: Called when the mouse moves within the Box while mouse buttons are pressed.
        Note that `start_dragging` and `stop_dragging` must be used to control whether
        on_drag events are listened to when using the base MouseBox.
    """

    on_press: Optional[MouseClickEventCallable] = None
    on_press_outside: Optional[MouseClickEventCallable] = None
    on_release: Optional[MouseClickEventCallable] = None
    on_hover: Optional[MouseMotionEventCallable] = None
    on_unhover: Optional[MouseMotionEventCallable] = None
    on_motion: Optional[MouseMotionEventCallable] = None
    on_drag: Optional[MouseDragEventCallable] = None


def do_nothing(*_, **__):
    """Do nothing on the event, and return True to mark the event as handled."""
    return EVENT_HANDLED


@dataclass(frozen=True)
class MouseVoidBoxDefinition(MouseBoxDefinition):
    """Definition of a mouse box that swallows all mouse events."""

    on_press: Optional[MouseClickEventCallable] = field(default=do_nothing)
    on_press_outside: Optional[MouseClickEventCallable] = field(default=do_nothing)
    on_release: Optional[MouseClickEventCallable] = field(default=do_nothing)
    on_hover: Optional[MouseMotionEventCallable] = field(default=do_nothing)
    on_unhover: Optional[MouseMotionEventCallable] = field(default=do_nothing)
    on_motion: Optional[MouseMotionEventCallable] = field(default=do_nothing)
    on_drag: Optional[MouseDragEventCallable] = field(default=do_nothing)


class MouseBox(ActiveBox):
    """
    A box that reacts to mouse events.

    Special care needs to be taken when trying to react to mouse drag events.
    By default, mouse press and release events do not start/stop dragging as you might expect,
    this is because it is impossible to define a generic behaviour that works for all situations.
    Therefore, you should call `start_dragging` and `stop_dragging` respectively when required.

    See `DraggableBox` for an example.
    """

    definition_type = MouseBoxDefinition

    def __init__(self, definition: MouseBoxDefinition):
        """
        Creates a new MouseBox.

        :param definition: Definition of the actions to take.
        """
        super(MouseBox, self).__init__(definition)
        self.definition: MouseBoxDefinition = definition

        # Whether the mouse is currently hovered over the Box or not.
        self._currently_hovered: bool = False

        # Used to record whether we're currently dragging this Box.
        # This is needed because if the user drags too fast then we can't just rely on
        # a drag event being inside the box area still.
        self._currently_dragging: bool = False

        # bitwise representation of pressed buttons, as pyglet defines them.
        self._currently_pressed: int = 0

    def _on_press(self, x: int, y: int, buttons: int, modifiers: int) -> Optional[bool]:
        """
        Called when the Box is clicked by the user.

        Calls the `on_press` callback from the definition, passing information about the event to
        the callback.
        """
        log.debug(f"Button {buttons} pressed on {self}.")
        if buttons is not None:
            self._currently_pressed = bitwise_add(self._currently_pressed, buttons)

        if self.definition.on_press is None:
            return None

        return self.definition.on_press(
            box=self, x=x, y=y, buttons=buttons, modifiers=modifiers
        )

    def _on_press_outside(
        self, x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        Called when a click occurs that does not intersect with this Box.

        Calls the `on_press_outside` callback from the definition, passing information about
        the event to the callback.
        """
        if self.definition.on_press_outside is None:
            return None

        return self.definition.on_press_outside(
            box=self, x=x, y=y, buttons=buttons, modifiers=modifiers
        )

    def _on_release(
        self, x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        Called when the mouse is released by the user within the button area.

        Not guaranteed to be called after every mouse click as the user may move the mouse off
        the Box before releasing.

        Calls the `on_release` callback from the definition, passing information about the event to
        the callback.
        """
        log.debug(f"Button {buttons} released on {self}.")

        if self.definition.on_release is not None:
            result = self.definition.on_release(
                box=self, x=x, y=y, buttons=buttons, modifiers=modifiers
            )
        else:
            result = None

        # Remove from currently selected after the on_release callback in case it is needed
        # in a sub class.
        if buttons is not None:
            self._currently_pressed = bitwise_remove(self._currently_pressed, buttons)

        return result

    def _on_hover(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """
        Called when the user moves the mouse over the Box without any mouse buttons pressed.

        Calls the `on_hover` callback from the definition, passing information about the event to
        the callback.
        """
        log.debug(f"Now hovering over {self}.")
        self._currently_hovered = True

        if self.definition.on_hover is None:
            return None

        return self.definition.on_hover(box=self, x=x, y=y, dx=dx, dy=dy)

    def _on_motion(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """
        Called when the user moves the mouse within the Box without any mouse buttons pressed.

        Calls the `on_motion` callback from the definition, passing information about the event to
        the callback.
        """
        if self.definition.on_motion is None:
            return None

        return self.definition.on_motion(box=self, x=x, y=y, dx=dx, dy=dy)

    def _on_unhover(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """
        Called when the user moves the mouse off the Box without any mouse buttons pressed.

        Calls the `on_unhover` callback from the definition, passing information about the event to
        the callback.
        """
        log.debug(f"No longer hovering over {self}.")
        # reset knowledge of which mouse buttons are pressed when mouse leaves the button.
        self._currently_pressed = 0
        self._currently_hovered = False

        if self.definition.on_unhover is None:
            return None

        return self.definition.on_unhover(box=self, x=x, y=y, dx=dx, dy=dy)

    def _on_drag(
        self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        Called when the mouse is moved by the user with a mouse button pressed.

        Calls the `on_drag` callback from the definition, passing information about the event to
        the callback.
        """
        if self.definition.on_drag is None:
            return None

        return self.definition.on_drag(
            box=self, x=x, y=y, dx=dx, dy=dy, buttons=buttons, modifiers=modifiers
        )

    def _should_handle_mouse_press(self, buttons: int) -> bool:
        """
        Determine if this Box should attempt to handle the mouse press event.

        :param buttons: Int indicating which mouse buttons are pressed (see pyglet).
        :return: True if this Box should handle the mouse click press.
        """
        return (
            self.definition.on_press is not None
            or self.definition.on_press_outside is not None
        )

    def _should_handle_mouse_release(self, buttons: int) -> bool:
        """
        Determine if this Box should attempt to handle the mouse release event.

        This checks if this Box was previously pressed with the same mouse button.
        This prevents the user clicking off the Box, moving onto it and then releasing.

        :param buttons: Int indicating which mouse buttons are pressed (see pyglet).
        :return: True if this Box should handle the mouse click release.
        """
        return self.definition.on_release is not None and bitwise_contains(
            self._currently_pressed, buttons
        )

    def _should_handle_mouse_hover(self) -> bool:
        """
        Determine if this Box should attempt to handle mouse hover and unhover events.

        :return: True if this Box should handle mouse hover and unhover events.
        """
        return (
            self.definition.on_hover is not None
            or self.definition.on_unhover is not None
        )

    def _should_handle_mouse_motion(self) -> bool:
        """
        Determine if this Box should attempt to handle mouse motion events.

        :return: True if this Box should handle mouse motion events.
        """
        return self.definition.on_motion is not None

    def _should_handle_mouse_drag(self) -> bool:
        """
        Determine if this Box should attempt to handle a mouse drag event.

        :return: True if this Box should handle the mouse drag event.
        """
        return self._currently_dragging and self.definition.on_drag is not None

    def on_mouse_press(
        self, x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        Cocos director callback when the mouse is pressed.

        Checks if the event happened in the area defined by this Box and, if so, handles it.
        """
        if not self._should_handle_mouse_press(buttons):
            return EVENT_UNHANDLED

        coord: Point2d = cocos.director.director.get_virtual_coordinates(x, y)
        if self.contains_coord(*coord):
            result = self._on_press(*coord, buttons, modifiers)
            if result is EVENT_HANDLED:
                log.debug(f"on_mouse_press consumed by {self}.")
                return EVENT_HANDLED
        else:
            self._on_press_outside(*coord, buttons, modifiers)

        return EVENT_UNHANDLED

    def on_mouse_release(
        self, x: int, y: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        Cocos director callback when the mouse is release.

        Checks if the event happened in the area defined by this Box and, if so, handles it.
        """
        if not self._should_handle_mouse_release(buttons):
            return EVENT_UNHANDLED

        coord: Point2d = cocos.director.director.get_virtual_coordinates(x, y)
        if self.contains_coord(*coord):
            result = self._on_release(*coord, buttons, modifiers)
            if result is EVENT_HANDLED:
                log.debug(f"on_mouse_release consumed by {self}.")
                return EVENT_HANDLED
        return EVENT_UNHANDLED

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> Optional[bool]:
        """
        Cocos director callback when the mouse is moved.

        Checks if the mouse moved onto or off the Box and triggers a hover or unhover event
        respectively.

        By default does not capture the event, so it may be handled by other entities as well.
        """
        if (
            not self._should_handle_mouse_hover()
            and not self._should_handle_mouse_motion()
        ):
            return EVENT_UNHANDLED

        result: Optional[bool] = None
        coord: Point2d = cocos.director.director.get_virtual_coordinates(x, y)
        if self.contains_coord(*coord):
            if not self._currently_hovered:
                result = self._on_hover(*coord, dx, dy)

            # If on_hover hasn't already handled the event, call on_motion.
            if result is not True:
                result = self._on_motion(*coord, dx, dy)

            # By default, do not return EVENT_HANDLED on hover as we could be
            # hovering over multiple things.
        elif self._currently_hovered:
            result = self._on_unhover(*coord, dx, dy)
            # By default, do not return EVENT_HANDLED on unhover as we have left
            # the button area.

        if result is EVENT_HANDLED:
            log.log(LTRACE, f"on_mouse_motion consumed by {self}.")
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def on_mouse_drag(
        self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """Cocos director callback when the mouse is moved while a mouse button is pressed."""
        if not self._should_handle_mouse_drag():
            return EVENT_UNHANDLED

        # We don't check for this event being within the bounds of the Box because we instead rely
        # on the setting of `self._currently_dragging` via another method (e.g. click/release) to
        # control whether drag events should be handled or not.
        coord: Point2d = cocos.director.director.get_virtual_coordinates(x, y)
        result = self._on_drag(*coord, dx, dy, buttons, modifiers)
        if result is EVENT_HANDLED:
            log.log(LTRACE, f"on_mouse_drag consumed by {self}.")
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def start_dragging(
        self, box: "MouseBox", x: int, y: int, buttons: int, modifiers: int,
    ) -> bool:
        """Callback to start dragging this Box."""
        log.debug(f"Start dragging {self}.")
        self._currently_dragging = True
        return EVENT_HANDLED

    def stop_dragging(
        self, box: "MouseBox", x: int, y: int, buttons: int, modifiers: int,
    ) -> bool:
        """Callback to stop dragging this Box."""
        log.debug(f"Stop dragging {self}.")
        self._currently_dragging = False
        return EVENT_HANDLED



"""Module of alignment definitions."""

from dataclasses import dataclass
from enum import Enum

import cocos


class HorizontalAlignment(Enum):
    """Enum of possible values for horizontal alignment in pyglet."""

    left = "left"
    right = "right"
    center = "center"


class VerticalAlignment(Enum):
    """Enum of possible values for vertical alignment in pyglet."""

    bottom = "bottom"
    center = "center"
    top = "top"


class VerticalTextAlignment(Enum):
    """Enum of possible values for vertical alignment of text in pyglet."""

    bottom = "bottom"
    center = "center"
    top = "top"
    # Baseline is the bottom of the first line of text, as opposed to `bottom` which is
    # the bottom of the pyglet text layout.
    baseline = "baseline"


class ZIndexEnum(Enum):
    """Indicators of where in the stack of cocos children to add a child."""

    top = "top"
    bottom = "bottom"


@dataclass(frozen=True)
class PositionalAnchor:
    """Pair of alignments that define a relative anchor point in 2D space."""

    horizontal: HorizontalAlignment
    vertical: VerticalAlignment

    @property
    def x(self) -> HorizontalAlignment:
        """Synonym for `self.horizontal`."""
        return self.horizontal

    @property
    def y(self) -> VerticalAlignment:
        """Synonym for `self.vertical`."""
        return self.vertical

    def get_coord_in_rect(self, width: int, height: int) -> cocos.draw.Point2:
        """
        Get the (x, y) coordinate of this anchor in the given rect.

        :param width: The width of the rect to consider.
        :param height: The height of the rect to consider.
        :return: cocos.draw.Point2 of the anchor position.
        """
        x, y = 0.0, 0.0

        if self.horizontal == HorizontalAlignment.left:
            x = 0
        elif self.horizontal == HorizontalAlignment.center:
            x = width / 2
        elif self.horizontal == HorizontalAlignment.right:
            x = width

        if self.vertical == VerticalAlignment.bottom:
            y = 0
        elif self.vertical == VerticalAlignment.center:
            y = height / 2
        elif self.vertical == VerticalAlignment.top:
            y = height

        return cocos.draw.Point2(x, y)


# Define the 9 basic anchor points of a rectangle.
LeftTop = PositionalAnchor(HorizontalAlignment.left, VerticalAlignment.top)
LeftCenter = PositionalAnchor(HorizontalAlignment.left, VerticalAlignment.center)
LeftBottom = PositionalAnchor(HorizontalAlignment.left, VerticalAlignment.bottom)
CenterTop = PositionalAnchor(HorizontalAlignment.center, VerticalAlignment.top)
CenterCenter = PositionalAnchor(HorizontalAlignment.center, VerticalAlignment.center)
CenterBottom = PositionalAnchor(HorizontalAlignment.center, VerticalAlignment.bottom)
RightTop = PositionalAnchor(HorizontalAlignment.right, VerticalAlignment.top)
RightCenter = PositionalAnchor(HorizontalAlignment.right, VerticalAlignment.center)
RightBottom = PositionalAnchor(HorizontalAlignment.right, VerticalAlignment.bottom)



"""Module defining ways for users to draw shapes."""

from abc import ABC
from dataclasses import dataclass, replace
from typing import Dict, Callable, Optional, Tuple

import pyglet

import cocos
from cocos.euclid import Point2, Vector2
from .mouse_box import MouseBox, MouseBoxDefinition
from ..components.box import BoxDefinition
from ..data_structures import ActiveGreen, Color


# Deliberately not frozen so we can update it as the rect is drawn.
@dataclass
class MouseDefinedRect:
    """
    Definition of a rect drawn by the mouse.

    Stores the coordinates of the dragged out rect, and which mouse buttons or keyboard
    modifiers were used at the start of the drawing.

    The minimum size is a 1x1 box. This is useful when the user performs a click
    without moving the mouse - you still get a non-zero rect which can intersect with other
    rects and/or be drawn as a single pixel.
    """

    # The drawing box that this rect was drawn in.
    canvas: "RectDrawingBox"
    buttons: int
    modifiers: int
    start_coord: Point2
    end_coord: Point2

    @property
    def vector(self) -> Vector2:
        """Vector from the start to the end coordinate."""
        return self.end_coord - self.start_coord

    @property
    def bottom_left(self) -> Point2:
        """
        The bottom-left point of the define rectangle.

        Accounts for rects drawn in any direction, i.e. it doesn't matter if the
        start coordinate is higher than the end coordinate.
        """
        return Point2(
            int(min(self.start_coord.x, self.end_coord.x)),
            int(min(self.start_coord.y, self.end_coord.y)),
        )

    @property
    def modifiers_as_string(self) -> str:
        """Return a string representing the modifiers used during drawing of this rect."""
        return pyglet.window.key.modifiers_string(self.modifiers)

    @property
    def dimensions(self) -> Tuple[int, int]:
        """Return the (width, height) of the defined rect, with a miniumum of (1, 1)."""
        move_vector = self.vector
        # Include the "or 1" to set the minimum size to be a 1x1 rect.
        # This both makes this a useful default to be a non-zero sized rect, but
        # also works around the cocos default of setting a rect to the full window size
        # if a dimension is 0.
        width, height = int(abs(move_vector[0])) or 1, int(abs(move_vector[1])) or 1
        return width, height

    def as_rect(self) -> cocos.rect.Rect:
        """Return the defined rect in coordinates relative to the canvas it was drawn on."""
        return cocos.rect.Rect(*self.bottom_left, *self.dimensions)

    def as_world_rect(self) -> cocos.rect.Rect:
        """Return the defined rect in world coordinates."""
        x, y = self.canvas.point_to_world(self.bottom_left)
        return cocos.rect.Rect(x, y, *self.dimensions)


@dataclass(frozen=True)
class RectDrawingBoxDefinition:
    """
    Definition of a box that the user can draw rectangles inside.

    Callbacks are given for start, change and completion of the drawing.

    The rect will be displayed during drawing using the defined color.
    """

    on_start: Optional[Callable[[MouseDefinedRect], None]] = None
    on_change: Optional[Callable[[MouseDefinedRect], None]] = None
    on_complete: Optional[Callable[[MouseDefinedRect], None]] = None
    color: Color = replace(ActiveGreen, a=100)


class DrawingBox(ABC, MouseBox):
    """Base class for defining boxes that can be drawn inside."""

    def __init__(self, definition: BoxDefinition):
        """
        Creates a new DrawableBox.

        :param definition: Definition of the shape of this Box.
        """
        definition = MouseBoxDefinition(
            width=definition.width,
            height=definition.height,
            on_press=self.on_press,
            on_release=self.on_release,
            on_drag=self.on_drag,
        )
        super(DrawingBox, self).__init__(definition)

    def on_press(
        self, box: MouseBox, x: int, y: int, buttons: int, modifiers: int,
    ) -> None:
        """Placeholder for subclasses to define what happens when a drawing begins."""
        pass

    def on_release(
        self, box: MouseBox, x: int, y: int, buttons: int, modifiers: int
    ) -> None:
        """Placeholder for subclasses to define what happens when a drawing is completed."""
        pass

    def on_drag(
        self,
        box: MouseBox,
        x: int,
        y: int,
        dx: int,
        dy: int,
        buttons: int,
        modifiers: int,
    ) -> None:
        """Placeholder for subclasses to define what happens when a drawing changes."""
        pass


class RectDrawingBox(DrawingBox):
    """
    A RectDrawingBox defines a rectangular area where Rects can be drawn using the mouse.

    The `on_complete` method in the definition is called every time a mouse button is
    released, returning the rect that was defined with that current set of buttons pressed.
    Multiple rects are recorded at once for each combination of mouse buttons that is pressed.

    Worked example:
        - Press left mouse button, start defining rect A at (L1, L2)
        - Move to (R1, R2)
        - Press right mouse button, start defining rect B at (R1, R2)
        - Move to (R3, R4)
        - Release right mouse button, call on_complete with rect defined by (R1, R2, R3, R4) with
            a MouseDefinedRect indicating that both left and right buttons are pressed.
        - Release left mouse button, call on_complete with rect defined by (L1, L2, R3, R4) with
            a MouseDefinedRect indicating that only the left button is pressed.
    """

    def __init__(
        self,
        definition: RectDrawingBoxDefinition,
        rect: Optional[cocos.rect.Rect] = None,
    ):
        """
        Creates a new RectDrawingBox.

        :param rect: Definition of the rectangle that this Box encompasses.
            If None, defaults to the entire window.
        """
        if rect is None:
            rect = cocos.rect.Rect(0, 0, *cocos.director.director.get_window_size())

        super(RectDrawingBox, self).__init__(rect)
        # Dict of rects currently being drawn.
        # Keys are the pyglet integers defining which mouse buttons are pressed.
        self._rects_in_progress: Dict[int, MouseDefinedRect] = {}
        self._rect_displays: Dict[int, cocos.layer.ColorLayer] = {}

        self.drawing_definition: RectDrawingBoxDefinition = definition

    def _update_rect_display(self, buttons: int) -> None:
        """Update the display of the rect during drawing."""
        rect = self._rects_in_progress[buttons].as_rect()
        layer = cocos.layer.ColorLayer(
            *self.drawing_definition.color.as_tuple_alpha(), rect.width, rect.height,
        )
        layer.position = rect.x, rect.y

        if buttons in self._rect_displays.keys():
            self.remove(self._rect_displays[buttons])

        self._rect_displays[buttons] = layer
        self.add(self._rect_displays[buttons])

    def on_press(
        self, box: MouseBox, x: int, y: int, buttons: int, modifiers: int,
    ) -> None:
        """
        Record the start of the rect drawing.

        This records the starting coordinate in local coordinates; the mouse buttons used
        and the keyboard modifiers used.

        Calls the `on_start` callback with the initial definition.
        """
        self._currently_dragging = True
        x, y = self.point_to_local((x, y))
        # End coord starts as the same as the start coord.
        self._rects_in_progress[buttons] = MouseDefinedRect(
            canvas=self,
            buttons=self._currently_pressed,
            modifiers=modifiers,
            start_coord=Point2(x, y),
            end_coord=Point2(x, y),
        )
        self._update_rect_display(buttons)

        if self.drawing_definition.on_start is not None:
            self.drawing_definition.on_start(self._rects_in_progress[buttons])

    def on_release(
        self, box: MouseBox, x: int, y: int, buttons: int, modifiers: int
    ) -> None:
        """
        Record the end of the drawing.

        Records the end coordinates and calls the `on_complete` callback with the final
        definition.
        """
        x, y = self.point_to_local((x, y))
        self._rects_in_progress[buttons].end_coord = Point2(x, y)

        if self.drawing_definition.on_complete is not None:
            self.drawing_definition.on_complete(self._rects_in_progress[buttons])

        del self._rects_in_progress[buttons]
        self.remove(self._rect_displays[buttons])
        del self._rect_displays[buttons]

        # Only stop dragging if every button has been released.
        if self._currently_pressed == 0:
            self._currently_dragging = False

    def on_drag(
        self,
        box: MouseBox,
        x: int,
        y: int,
        dx: int,
        dy: int,
        buttons: int,
        modifiers: int,
    ) -> None:
        """
        Update the definition of the rect being drawn with the latest mouse position.

        Updates the visible rect to the latest rect definition.

        Calls the `on_change` callback with the latest definition.
        """
        x, y = self.point_to_local((x, y))

        # The "buttons" parameter on drag is the bitwise sum of all currently pressed buttons.
        # We already have a record of what's pressed in _rects_in_progress, so just update them all.
        for button, rectdefn in self._rects_in_progress.items():
            rectdefn.end_coord = Point2(x, y)
            self._update_rect_display(button)

            if self.drawing_definition.on_change is not None:
                self.drawing_definition.on_change(rectdefn)



"""Data structures to aid definition of graphics objects."""

from dataclasses import dataclass
from enum import Enum
from typing import Tuple, List, Any, Dict


@dataclass(frozen=True)
class Color:
    """Definition of an RGB + alpha color."""

    r: int
    g: int
    b: int
    a: int = 255

    def as_tuple(self) -> Tuple[int, int, int]:
        """Return the RGB values as a tuple."""
        return self.r, self.g, self.b

    def as_tuple_alpha(self) -> Tuple[int, int, int, int]:
        """Return the RGB + alpha values as a tuple."""
        return self.r, self.g, self.b, self.a


White = Color(255, 255, 255)
ActiveGreen = Color(0, 200, 255)
PassiveBlue = Color(0, 120, 255)
ActiveBlue = Color(0, 200, 255)
MutedBlue = Color(0, 80, 255)

Black = Color(0, 0, 0)
LightGrey = Color(160, 160, 190)
Grey = Color(130, 130, 170)
DarkGrey = Color(100, 100, 150)


def shimmer_dict_factory(items: List[Tuple[Any, Any]]) -> Dict[Any, Any]:
    """
    Custom dict factory for use with `as_dict` on shimmer dataclasses.

    Converts the following fields:
        - Enums to the corresponding value.
    """
    _dict = dict(items)
    for key, value in _dict.items():
        if isinstance(value, Enum):
            _dict[key] = value.value
    return _dict



"""Module defining text fonts."""
from dataclasses import dataclass, asdict
from typing import Optional, Dict, Any

from pyglet.font import load as load_font

from shimmer.display.data_structures import Color, White, shimmer_dict_factory


@dataclass(frozen=True)
class FontDefinition:
    """Definition of a text Font. See `pyglet.text.Label` for full parameter descriptions."""

    font_name: str
    font_size: int
    bold: bool = False
    italic: bool = False
    color: Color = White
    dpi: Optional[int] = None  # Resolution of the font. Defaults to 96.

    def to_pyglet_label_kwargs(self) -> Dict[str, Any]:
        """Convert this definition into cocos/pyglet compatible kwargs."""
        _dict = asdict(self, dict_factory=shimmer_dict_factory)
        # We want `color` to be a tuple rather than a dict.
        _dict["color"] = self.color.as_tuple_alpha()
        return _dict

    @property
    def height(self) -> int:
        """
        Discover the height of this Font.

        This is a relatively expensive call, so cache it if you need to use it often.
        """
        font = load_font(
            self.font_name, self.font_size, self.bold, self.italic, self.dpi
        )
        return font.ascent - font.descent


Calibri = FontDefinition("calibri", 16)
ComicSans = FontDefinition("comic sans", 16)



"""Common methods for programmable code blocks."""

from dataclasses import dataclass
from inspect import signature
from typing import Set, Callable, Generator, Optional

from .definition import Program


@dataclass
class Programmable:
    """
    Definition of a user-editable program.

    The `methods` are the valid methods that the user can choose to use.

    The `program` contains the actual program definition.
    """

    methods: Set[Callable]
    program: Optional[Program] = None

    @property
    def bool_methods(self) -> Generator[Callable, None, None]:
        """
        Return the set of `methods` that return a bool value.

        This is used to only present boolean methods to users for use in If/Elif/While statements.
        """
        for method in self.methods:
            if issubclass(signature(method).return_annotation, bool):
                yield method



"""Visual display of a code block Instruction."""

import logging
from dataclasses import dataclass, replace
from typing import Optional

import cocos
from shimmer.display.components.draggable_box import (
    DraggableBox,
    DraggableBoxDefinition,
)
from shimmer.display.data_structures import Color, ActiveGreen
from shimmer.display.primitives import create_color_rect
from shimmer.display.widgets.button import ButtonDefinition, Button
from shimmer.engine.programmable.definition import Instruction

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class InstructionDisplayDefinition(ButtonDefinition):
    """
    Definition of the visual display of an Instruction.

    If text is None, then it will be set to the formatted `method_str` of the instruction.
    """

    # Color to display over the Instruction button while the instruction is being executed.
    while_executing_mask: Optional[Color] = replace(ActiveGreen, a=100)
    # String format to use. Will be formatted with `instruction.method_str()`.
    text_format: str = "{}"
    # Whether the instruction should be draggable.
    draggable: bool = False
    # The width of the draggable anchor for this instruction.
    draggable_width: int = 20


class InstructionDisplay(Button):
    """
    Display for a code block Instruction.

    Comprised of 3 Boxes:
      - A Button - e.g. for allowing the user to edit the instruction.
      - A drag anchor aligned on the left of the button.
      - A color mask that overlays the button while the instruction is being executed.
    """

    def __init__(
        self, instruction: Instruction, definition: InstructionDisplayDefinition
    ):
        """
        Create an InstructionDisplay.

        :param instruction: The Instruction to display.
        :param definition: Definition of the display of the Instruction.
        """
        button_text = self.get_button_text(instruction, definition)
        definition = replace(definition, text=button_text)
        self.instruction = instruction

        self.instruction.on_execute_start = self.show_mask
        self.instruction.on_execute_complete = self.hide_mask

        self.drag_anchor: Optional[DraggableBox] = None
        self.executing_mask: Optional[cocos.layer.ColorLayer] = None

        super(InstructionDisplay, self).__init__(definition)
        self.definition: InstructionDisplayDefinition = definition
        self.update_draggable_anchor()
        self.update_mask()

    def on_size_change(self):
        """Update children as needed if the size of this Box changes."""
        super(InstructionDisplay, self).on_size_change()
        self.update_mask()

    @staticmethod
    def get_button_text(
        instruction: Instruction, definition: InstructionDisplayDefinition
    ) -> Optional[str]:
        """
        Get the text of the button should display.

        By default it will be `instruction.method_str()` applied to the defined text_format.
        """
        if definition.text is None:
            return definition.text_format.format(instruction.method_str())
        else:
            return None

    def update_draggable_anchor(self):
        """Re-create the draggable anchor."""
        if self.drag_anchor is not None:
            self.remove(self.drag_anchor)

        if self.definition.draggable is False:
            self.drag_anchor = None
            return

        self.drag_anchor = DraggableBox(
            DraggableBoxDefinition(
                width=self.definition.draggable_width, height=self.rect.height
            )
        )
        self.add(self.drag_anchor, z=1)

    def update_mask(self):
        """Re-create the color mask, starting off invisible."""
        if self.executing_mask is not None:
            self.remove(self.executing_mask)

        if self.definition.while_executing_mask is None:
            self.executing_mask = None
            return

        # Create the mask invisible (with 0 opacity), it should be made visible using the
        # instruction callbacks.
        self.executing_mask = create_color_rect(
            width=self.rect.width,
            height=self.rect.height,
            color=replace(self.definition.while_executing_mask, a=0),
        )
        self.add(self.executing_mask, z=1)

    def show_mask(self):
        """Make the color mask visible."""
        if (
            self.executing_mask is not None
            and self.definition.while_executing_mask is not None
        ):
            self.executing_mask.opacity = self.definition.while_executing_mask.a

    def hide_mask(self):
        """Make the color mask invisible."""
        if self.executing_mask is not None:
            self.executing_mask.opacity = 0



"""
Definition of the Box primitive.

A Box is a CocosNode that defines an area.
"""

import logging
from dataclasses import dataclass, fields, replace
from typing import Optional, Type, Iterable, Tuple, Union

import cocos
from ..alignment import (
    ZIndexEnum,
    PositionalAnchor,
    CenterCenter,
)
from ..data_structures import Color
from ..primitives import create_color_rect

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class BoxDefinition:
    """Common definition to all Boxes."""

    # If either `width` or `height` are None then the box size will dynamically encompass all of
    # its children in the both dimensions.
    # Set to 0 to match the entire window size.
    width: Optional[int] = None
    height: Optional[int] = None

    background_color: Optional[Color] = None

    def __str__(self) -> str:
        """String representation that excludes fields that have their default value."""
        params = {}
        for field in fields(self):
            value = getattr(self, field.name)
            if hasattr(value, "__name__"):
                # Don't recurse into getting the string representation of methods, as these
                # are often bound methods and result in calling back into this function.
                value = value.__name__

            if value != field.default:
                params[field.name] = str(value)
        return f"{self.__class__.__name__}({params})"

    @property
    def is_dynamic_sized(self) -> bool:
        """True if the defined Box should have dynamic size, otherwise False."""
        return self.width is None or self.height is None


class Box(cocos.cocosnode.CocosNode):
    """A CocosNode that has a defined rectangular area."""

    definition_type: Type[BoxDefinition] = BoxDefinition

    def __init__(self, definition: Optional[BoxDefinition] = None):
        """Creates a new Box."""
        super(Box, self).__init__()
        if definition is None:
            definition = self.definition_type()

        self.definition = definition
        self._width: int = self.definition.width or 0
        self._height: int = self.definition.height or 0
        self._rect = cocos.rect.Rect(0, 0, self._width, self._height)
        self._background: Optional[cocos.layer.ColorLayer] = None
        self.update_background()

    def __repr__(self):
        """String representation of this Box."""
        return f"{self.__class__.__name__}({self.definition})"

    def __str__(self):
        """String representation of this Box."""
        return (
            f"[{id(self)}][{self.world_rect}] {repr(self)} "
            f"child of <{repr(self.parent)}>"
        )

    def contains_coord(self, x: int, y: int) -> bool:
        """Returns whether the point (x,y) is inside the box."""
        p = self.point_to_local((x, y))
        return self.rect.contains(p.x, p.y)

    @property
    def rect(self) -> cocos.rect.Rect:
        """
        Return the rect that this box encompasses.

        This rect is in the parent coordinate space.
        """
        return self._rect

    @property
    def world_rect(self) -> cocos.rect.Rect:
        """Get the rect for this Box in world coordinate space."""
        return cocos.rect.Rect(
            *self.point_to_world((0, 0)), self.rect.width, self.rect.height
        )

    def update_rect(self):
        """
        Update the cached rect definition.

        If the rect size changes, then `on_size_change` will be called.
        """
        if self.definition.is_dynamic_sized:
            children_rect = self.bounding_rect_of_children()
            self._width = children_rect.width
            self._height = children_rect.height
        else:
            self._width = self.definition.width or 0
            self._height = self.definition.height or 0

        if self._width != self._rect.width or self._height != self._rect.height:
            self._rect.set_size((self._width, self._height))
            self.on_size_change()

    def on_size_change(self):
        """
        Called when the size of the Box changes.

        Only applicable for dynamically sized Boxes.

        Called when `update_rect` detects a size change.
        """
        self.update_background()

    def add(
        self,
        child: cocos.cocosnode.CocosNode,
        z: int = 0,
        name: Optional[str] = None,
        no_resize: bool = False,
    ) -> None:
        """
        Add a child to this Box. See CocosNode for parameter arguments.

        When adding a child, check for updating dynamic size of this Box.

        :param child: CocosNode to add.
        :param z: See CocosNode
        :param name: See CocosNode
        :param no_resize: If True, then the size of this box is not dynamically changed.
        """
        super(Box, self).add(child, z, name)
        if self.definition.is_dynamic_sized and not no_resize:
            self.update_rect()

    def remove(self, child: cocos.cocosnode.CocosNode, no_resize: bool = False) -> None:
        """
        Remove a child from this Box. See CocosNode for parameter arguments.

        When removing a child, check for updating dynamic size of this Box.

        :param child: CocosNode to remove.
        :param no_resize: If True, then the size of this box is not dynamically changed.
        """
        super(Box, self).remove(child)
        if self.definition.is_dynamic_sized and not no_resize:
            self.update_rect()

    def get_z_value(self) -> Optional[int]:
        """
        Get the z value of this Box in its parents children list.

        Returns None if this child has no parent.

        Raises ValueError if this child is not a child of parent (should never happen).
        """
        if self.parent is None:
            return None

        for z, child in self.parent.children:
            if child is self:
                return z
        raise ValueError(f"{self} is not a child of its parent.")

    def set_z_value(self, z: Union[int, ZIndexEnum]) -> None:
        """
        Set the z value of this box in its parents list of children.

        The z value controls ordering of drawing, and ordering of event handling.

        Higher z means:
          - Drawn on top of lower z nodes.
          - Receives events first (and therefore may stop them propagating to lower z nodes)
                Note: This happens because in ActiveBox we push own events before children do.

        Warning: This removes and re-adds this node to the parent, so `on_exit` and `on_enter`
            are called.
            For ActiveBoxes this will remove and re-add it to the event stack. So if the z value
            is changed during an event callback, the event may skip children of THIS node.
            This can be worked around by consuming the event when changing the Z value, and
            re-dispatching it (see FocusBox for an example of this).

        :param z: Int or member of ZIndexEnum to set the z value to.
        """
        if z is ZIndexEnum.top:
            # Children are in z order from lowest to highest, so highest z is the last one.
            # Children are stored as the tuple (z, node)
            if self.parent.children[-1][1] is self:
                # If this box is already at the top, then do nothing.
                return
            z = self.parent.children[-1][0] + 1
        elif z is ZIndexEnum.bottom:
            # Children are in z order from lowest to highest, so lowest z is the first one.
            # Children are stored as the tuple (z, node)
            if self.parent.children[0][1] is self:
                # If this box is already at the top, then do nothing.
                return
            z = self.parent.children[0][0] - 1

        # Remove and re-add to insert this node in the child list correctly, and perform
        # `on_exit` and `on_enter` to make sure that the node is in the correct place in the
        # event stack.
        self.parent.remove(self)
        self.parent.add(self, z=z)

    def update_background(self) -> None:
        """
        Re-create the background of this Box to take account of changed size or color.

        If the background color is None, then the background is removed.
        """
        # Remove the old background
        if self._background is not None:
            self.remove(self._background, no_resize=True)

        if self.definition.background_color is None:
            return

        if self._width > 0 and self._height > 0:
            self._background = create_color_rect(
                self._width, self._height, self.definition.background_color
            )
            self._background.position = (0, 0)
            self.add(self._background, z=-100, no_resize=True)

    def _set_background_color(self, color: Color) -> None:
        """Set the color of the background of this box to the given Color."""
        self.definition = replace(self.definition, background_color=color)
        if self._background is not None:
            self._background.color = color.as_tuple()
            self._background.opacity = color.a
        else:
            self.update_background()

    def get_coordinates_of_anchor(self, anchor: PositionalAnchor) -> cocos.draw.Point2:
        """
        Get the (x, y) coordinate of the given anchor point in this Box.

        :param anchor: The anchor to get the position of.
        :return: cocos.draw.Point2 of the anchor position.
        """
        return anchor.get_coord_in_rect(self._width, self._height)

    def vector_between_anchors(
        self,
        other: "Box",
        other_anchor: PositionalAnchor,
        self_anchor: PositionalAnchor,
    ) -> cocos.draw.Vector2:
        """
        Get the vector between the anchor of this box, and the anchor of the other box.

        See `align_anchor_with_other_anchor` for argument descriptions.
        """
        if not self.is_running:
            # We need to work in a common coordinate space. The easiest choice is the director.
            # We could check for the director as an ancestor; but it is sufficient to check that
            # the Box is running (i.e. in the current scene).
            #
            # If this Box is not in the scene, then we cannot calculate relative positioning
            # reliably. However, it still works accurately for Boxes that are being aligned
            # with their parent or with siblings. Therefore only log rather than error out.
            log.debug(
                f"This Box is not in the scene, "
                f"so setting relative positioning may by unstable. {self=}, {other=}"
            )

        # Get the anchor coordinates in each Boxes coordinate space.
        self_point = self.get_coordinates_of_anchor(self_anchor)
        other_point = other.get_coordinates_of_anchor(other_anchor)

        # Translate those coordinates into the world coordinate space to ensure a common
        # coordinate space.
        self_world_point = self.point_to_world(self_point)
        other_world_point = other.point_to_world(other_point)

        # Get the difference between those two points in world space, which is the amount we need
        # to translate this Box by to align the two anchors.
        anchor_vector = other_world_point - self_world_point
        return anchor_vector

    def align_anchor_with_other_anchor(
        self,
        other: "Box",
        other_anchor: PositionalAnchor = CenterCenter,
        self_anchor: Optional[PositionalAnchor] = None,
        spacing: Union[int, Tuple[int, int], cocos.draw.Point2] = 0,
    ) -> None:
        """
        Set an anchor of this Box to be aligned an anchor of the other Box.

        For example, to align the right edge of this box with the left edge of the other:

            self.align_anchor_with_other_anchor(other, RightCenter, LeftCenter)

        :param other: The Box to align this one with.
        :param other_anchor: The anchor point of the other Box to align.
            Defaults to CenterCenter.
        :param self_anchor: The anchor point of this Box to align.
            Defaults to match `other_anchor`.
        :param spacing: Pixels to leave between the anchors.
            If an `int` is given, the direction of the spacing the vector defined between the
                given self_anchor and the CenterCenter of this Box.
                Positive integers serve to make more space between the edges of the two Boxes,
                while negative integers serve to cause the Box edges to overlap.
            If a Tuple[int, int] is given, then it is treated as an exact (x, y) offset.
        """
        if self_anchor is None:
            self_anchor = other_anchor

        anchor_vector = self.vector_between_anchors(other, other_anchor, self_anchor)

        # Now account for spacing.
        # If the spacing is an integer, then it is applied in the direction defined from
        # the given self_anchor to the CenterCenter of this Box.
        if isinstance(spacing, (cocos.draw.Point2, cocos.draw.Vector2)):
            spacing_vector = spacing
        elif isinstance(spacing, tuple):
            spacing_vector = cocos.draw.Point2(*spacing)
        else:
            # spacing is an int
            if spacing == 0 or self_anchor == CenterCenter:
                spacing_vector = cocos.draw.Point2(0, 0)
            else:
                spacing_vector = self.get_coordinates_of_anchor(
                    CenterCenter
                ) - self.get_coordinates_of_anchor(self_anchor)
                spacing_vector.normalize()
                spacing_vector *= spacing

        self.position += anchor_vector + spacing_vector

    def bounding_rect_of_children(self) -> cocos.rect.Rect:
        """
        Return the rect that contains all the child (and grandchild etc.) Boxes of this Box.

        The Rect of this Box is not taken into account.

        The bottom-left corner of the returned rect is aligned with the origin of this Box.
        """

        def get_world_rect(
            node: cocos.cocosnode.CocosNode,
        ) -> Optional[cocos.rect.Rect]:
            if isinstance(node, Box) and node is not self:
                return node.world_rect
            return None

        world_rects = list(filter(lambda x: x is not None, self.walk(get_world_rect)))
        if len(world_rects) != 0:
            bounding_world_rect = bounding_rect_of_rects(world_rects)
            bounding_local_rect = cocos.rect.Rect(
                *self.point_to_local(bounding_world_rect.position),
                bounding_world_rect.width,
                bounding_world_rect.height,
            )
        else:
            bounding_local_rect = cocos.rect.Rect(0, 0, 0, 0)

        return bounding_local_rect


class ActiveBox(Box):
    """A Box that adds itself to the cocos director event handlers when entering the scene."""

    def on_enter(self):
        """Called every time just before the node enters the stage."""
        # Deliberately push handlers for this node before calling for sub nodes. This causes
        # handlers for children to be inserted higher in the stack than their parents,
        # resulting in them getting higher priority on handling events.
        # This makes sense because children are generally smaller / drawn on top of parents.
        cocos.director.director.window.push_handlers(self)
        super(ActiveBox, self).on_enter()

    def on_exit(self):
        """Called every time just before the node exits the stage."""
        cocos.director.director.window.remove_handlers(self)
        super(ActiveBox, self).on_exit()


def bounding_rect_of_rects(rects: Iterable[cocos.rect.Rect]) -> cocos.rect.Rect:
    """Return the minimal rect needed to cover all the given rects."""
    lefts, rights, tops, bottoms = zip(
        *((rect.x, rect.x + rect.width, rect.y + rect.height, rect.y) for rect in rects)
    )

    left = min(lefts)
    bottom = min(bottoms)
    right = max(rights)
    top = max(tops)

    x = min(left, right)
    y = min(bottom, top)
    width = max(left, right) - x
    height = max(bottom, top) - y

    return cocos.rect.Rect(x, y, width, height)



"""Definition of a Gauge."""

from dataclasses import dataclass
from typing import Any

from .base import FloatRange


@dataclass
class GaugeDefinition:
    """
    Definition of a value-constrained indicator.

    The value of the gauge is constrained to within `full_range`.

    `high_range` and `low_range` are subsections of the gauge used to indicate optimal values
     of the gauge.
    """

    full_range: FloatRange
    low_range: FloatRange
    high_range: FloatRange
    value: float = 0

    @property
    def low_fraction_start(self) -> float:
        """Fractional position of lowest point of the low range in the full range."""
        return self.low_range.min / self.full_range.len()

    @property
    def high_fraction_start(self) -> float:
        """Fractional position of lowest point of the high range in the full range."""
        return self.high_range.min / self.full_range.len()

    @property
    def value_fraction_start(self) -> float:
        """Fractional position of the value in the full range."""
        return self.value / self.full_range.len()

    @property
    def low_fraction_len(self) -> float:
        """Fraction of the full range that the "low_range" covers."""
        return self.low_range.len() / self.full_range.len()

    @property
    def high_fraction_len(self) -> float:
        """Fraction of the full range that the "high_range" covers."""
        return self.high_range.len() / self.full_range.len()

    def __setattr__(self, key: str, value: Any) -> None:
        """Constrain the value of the gauge to within its full_range."""
        if key == "value":
            if value < self.full_range.min:
                value = self.full_range.min
            elif value > self.full_range.max:
                value = self.full_range.max
        super(GaugeDefinition, self).__setattr__(key, value)



"""Definition of ways to arrange Boxes."""

from abc import abstractmethod
from dataclasses import dataclass
from typing import List, Union, Optional

import cocos
from .box import Box


@dataclass(frozen=True)
class BoxLayoutDefinition:
    """
    Definition of a layout of Boxes.

    Supported layouts are:
      - vertically column
      - horizontal row
      - rectangular grid

    :param spacing: Number of pixels to leave between Boxes.
    :param boxes_per_row: If given, boxes will fill row-by-row upwards; up to a maximum of `height`.
    :param boxes_per_column: If given and `boxes_per_row` is not, then boxes will fill
        column-by-column from left to right with no max width.

        Examples of values of (boxes_per_row, boxes_per_column):
          - (None, 1) results in a single row of buttons
          - (1, None) results in a single column of buttons
          - (x, y)    results in a x by y grid of buttons
    """

    spacing: int = 10
    boxes_per_row: Optional[int] = None
    boxes_per_column: Optional[int] = 1


class BoxLayout(Box):
    """A collection of Boxes with a well defined layout."""

    def __init__(self, boxes: List[Box], spacing: int = 10):
        """
        Create a new BoxLayout.

        :param boxes: List of Boxes to include in the layout.
        :param spacing: Number of pixels to leave between Boxes.
        """
        super(BoxLayout, self).__init__()
        self._boxes = []  # type: List[Box]
        self._spacing = spacing
        for box in boxes:
            self.add(box)
        self.update_layout()

    @property
    def spacing(self) -> int:
        """Return the spacing between items in the layout."""
        return self._spacing

    @spacing.setter
    def spacing(self, value: int) -> None:
        """Set the spacing between items in the layout, and update the layout."""
        self._spacing = value
        self.update_layout()

    def remove(
        self, obj: Union[cocos.cocosnode.CocosNode, Box], no_resize: bool = False
    ) -> None:
        """
        Remove an object from this Layout, and update the position of other Boxes if needed.

        :param obj: CocosNode to remove.
        :param no_resize: If True, then the size of this box is not dynamically changed.
        """
        super(BoxLayout, self).remove(obj, no_resize=no_resize)
        if isinstance(obj, Box):
            self._boxes.remove(obj)
            self.update_layout()

    def add(
        self,
        child: Union[cocos.cocosnode.CocosNode, Box],
        z: int = 0,
        name: Optional[str] = None,
        position: Optional[int] = None,
    ) -> None:
        """
        Add a child cocosnode. If it's a Box then include it in the layout.

        :param child: CocosNode to add.
        :param z: See CocosNode
        :param name: See CocosNode
        :param position: Index to insert the box into the list of boxes. Defaults to the end.
        """
        super(BoxLayout, self).add(child, z, name)
        if isinstance(child, Box):
            if position is None:
                self._boxes.append(child)
            else:
                self._boxes.insert(position, child)
            self.update_layout()

    @abstractmethod
    def update_layout(self) -> None:
        """Update the position of all boxes in this Layout."""
        pass


class BoxRow(BoxLayout):
    """Arranges boxes horizontally. Boxes are arranged from left to right."""

    def update_layout(self) -> None:
        """Update the position of all boxes in this Layout."""
        # Don't update layout if there are no boxes to layout.
        if not self._boxes:
            return

        x_total = 0
        for box in self._boxes:
            box.x = x_total
            x_total += box.rect.width + self._spacing

        self.update_rect()


class BoxColumn(BoxLayout):
    """Arranges boxes vertically. Boxes are arranged from bottom to top."""

    def update_layout(self) -> None:
        """Update the position of all boxes in this Layout."""
        # Don't update layout if there are no boxes to layout.
        if not self._boxes:
            return

        y_total = 0
        for box in self._boxes:
            box.y = y_total
            y_total += box.rect.height + self._spacing

        self.update_rect()


def build_rectangular_grid(
    boxes: List[Box],
    width: Optional[int] = None,
    height: Optional[int] = None,
    spacing: int = 10,
) -> Union[BoxRow, BoxColumn]:
    """
    Build a rectangular grid of boxes.

    Fills from bottom-left.

    Return type is either a BoxRow of BoxColumns; or a BoxColumn of BoxRows.

    :param boxes: Boxes to include.
    :param width: If given, boxes will fill row-by-row upwards; up to a maximum of `height`.
    :param height: If given and `width` is not, then boxes will fill column-by-column from
        left to right with no max width.
    :param spacing: The spacing between the boxes.
    :return: Returns a BoxRow or BoxColumn depending on the stacking order.
    """
    if width is not None:
        grid_type = BoxColumn
        element_type = BoxRow
        boxes_per_element = width
        max_elements = height
    elif height is not None:
        grid_type = BoxRow
        element_type = BoxColumn
        boxes_per_element = height
        max_elements = width
    else:
        raise ValueError(
            "Grid layout is undefined with both `width` and `height` undefined."
        )

    # Maximum elements is one per box
    max_elements = max_elements if max_elements is not None else len(boxes)
    element_index = 0
    elements = []
    while True:
        if element_index >= max_elements:
            # Reached max requested width/height
            break

        box_batch = boxes[
            element_index * boxes_per_element : (element_index + 1) * boxes_per_element
        ]

        if not box_batch:
            # We've run out of boxes.
            # Last element might have fewer than other elements, but that's ok.
            break

        elements.append(element_type(box_batch, spacing=spacing))
        element_index += 1

    return grid_type(elements, spacing=spacing)


def create_box_layout(
    definition: BoxLayoutDefinition, boxes: List[Box]
) -> Union[BoxRow, BoxColumn]:
    """
    Create a layout of boxes based on the given definition.

    :param definition: BoxLayoutDefinition to use.
    :param boxes: Boxes to include in the layout.
    :return: BowRow or BoxColumn containing the given boxes.
    """
    width_height = definition.boxes_per_row, definition.boxes_per_column
    if width_height == (None, 1):
        return BoxRow(boxes, definition.spacing)
    elif width_height == (1, None):
        return BoxColumn(boxes, definition.spacing)
    else:
        return build_rectangular_grid(
            boxes,
            definition.boxes_per_row,
            definition.boxes_per_column,
            definition.spacing,
        )



"""Useful methods for enhancing logging."""

import logging

from functools import update_wrapper
from typing import Callable

LTRACE: int = 5

logging.addLevelName(LTRACE, "TRACE")


def log_exceptions(logger: logging.Logger) -> Callable[[Callable], Callable]:
    """
    Decorator that logs any exceptions raised by the decorated method.

    Useful for debugging callbacks from user input (e.g. any pyglet event) because otherwise they
    silently fail.

    :param logger: The Logger to use to log the exception.
    """

    def outer(func):
        """The actual decorator."""

        def wrapper(*args, **kwargs):
            """The wrapper for the function."""
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logger.exception(e)

        wrapped = update_wrapper(wrapper, func)
        return wrapped

    return outer



"""
A set of buttons that users can choose one, or many of.

For example, this can be used to create:
- a radio button
- a "select all that apply" set of buttons
"""

from collections import defaultdict
from dataclasses import dataclass, field, replace
from typing import Set, Optional, Union, Dict, Any

from shimmer.display.helpers import bundle_callables
from .button import ButtonDefinition, ToggleButton
from .question_definition import MultipleChoiceQuestionDefinition
from ..components.box import Box, BoxDefinition
from ..components.box_layout import (
    BoxColumn,
    BoxRow,
    create_box_layout,
    BoxLayoutDefinition,
)
from ..components.mouse_box import MouseClickEventCallable


@dataclass(frozen=True)
class MultipleChoiceButtonsDefinition(BoxDefinition):
    """Definition of a set of multiple choice buttons."""

    question: MultipleChoiceQuestionDefinition = MultipleChoiceQuestionDefinition(
        text=""
    )

    # The style of the buttons to display. The text field will be ignored.
    button: ButtonDefinition = field(default_factory=ButtonDefinition)

    # How to arrange the multiple buttons.
    layout: BoxLayoutDefinition = field(default_factory=BoxLayoutDefinition)


class MultipleChoiceButtons(Box):
    """A set of buttons that together represent multiple options to the user."""

    def __init__(self, definition: MultipleChoiceButtonsDefinition):
        """Create a MultipleChoiceButtons."""
        super(MultipleChoiceButtons, self).__init__(definition)
        self.definition: MultipleChoiceButtonsDefinition = definition
        self._layout: Optional[Union[BoxRow, BoxColumn]] = None
        self._buttons: Dict[Union[str, Box], ToggleButton] = {}
        self._current_selection: Dict[Union[str, Box], bool] = defaultdict(bool)
        self.update_layout()
        self.set_to_defaults()

    @property
    def currently_selected(self) -> Set[Union[str, Box]]:
        """Return a list of the currently selected options."""
        return set(
            item[0]
            for item in filter(
                lambda item: item[1] is True, self._current_selection.items()
            )
        )

    def set_to_defaults(self) -> None:
        """
        Set all toggle buttons to match the defined default state.

        Will call each buttons callback if their state changes.
        """
        for choice, button in self._buttons.items():
            if choice in self.definition.question.defaults:
                button.is_toggled = True
            else:
                button.is_toggled = False

    def _create_choice_callback_wrapper(
        self,
        chosen: Union[str, Box],
        existing_on_press: Optional[MouseClickEventCallable],
    ) -> MouseClickEventCallable:
        """
        Create callback to notify this multiple choice of a button press.

        Also calls the existing on_select method if it exists.

        :param chosen: The choice represented by the button.
        :param existing_on_press: Optional on_select callback that already is defined on the button.
        """

        def inner(box: ToggleButton, *_: Any, **__: Any) -> None:
            self._handle_button_select_or_deselect(chosen, box)

        if existing_on_press:
            return bundle_callables(inner, existing_on_press)
        else:
            return inner

    def _handle_button_select_or_deselect(
        self, chosen: Union[str, Box], button: ToggleButton
    ) -> None:
        """
        Called when a child button is selected.

        :param chosen: The choice represented by the button.
        :param button: The button that was selected.
        """
        self._current_selection[chosen] = button.is_toggled

        if self.definition.question.on_select is not None:
            self.definition.question.on_select(
                self.currently_selected, chosen, button.is_toggled
            )

        # If only 1 button is allowed to be selected, deselect all other ones.
        # This will actually call back into here for each deselection as they will call their
        # toggle change callback. However, we limit recursion based on whether the button has been
        # set to True, which every other button won't be.
        if not self.definition.question.allow_multiple and button.is_toggled:
            for choice, button in self._buttons.items():
                if choice != chosen:
                    button.is_toggled = False

    def _create_buttons(self):
        """Create the set of buttons for each choice."""
        buttons = []
        for choice in self.definition.question.choices:
            if isinstance(choice, str):
                text = choice
            else:
                text = ""

            defn = replace(
                self.definition.button,
                text=text,
                on_press=self._create_choice_callback_wrapper(
                    choice, self.definition.button.on_press
                ),
                on_release=self._create_choice_callback_wrapper(
                    choice, self.definition.button.on_release
                ),
            )
            button = ToggleButton(defn)
            if isinstance(choice, Box):
                button.add(choice, z=-100)
            buttons.append(button)
        self._buttons = dict(zip(self.definition.question.choices, buttons))
        return buttons

    def update_layout(self):
        """Recreate the buttons and the layout of this multiple choice."""
        if self._layout is not None:
            self.remove(self._layout)

        self._layout = create_box_layout(self.definition.layout, self._create_buttons())
        self.add(self._layout)



"""Modules defining methods common to widgets."""

from dataclasses import dataclass


@dataclass
class FloatRange:
    """Definition of a range of floats."""

    min: float
    max: float

    def contains(self, value: float) -> bool:
        """Return True if the value is within the min/max (inclusive) of this range."""
        return (value <= self.max) and (value >= self.min)

    def __contains__(self, item: float) -> bool:
        """Return True if the value is within the min/max (inclusive) of this range."""
        return self.contains(item)

    def len(self) -> float:
        """Returnsthe difference between max and min."""
        return self.max - self.min



"""An Entitiy that advances through various stages."""

import asyncio
import logging
from typing import Optional

from .definition import StagedEntityDefinition
from ..entity.entity import Entity

log = logging.getLogger(__name__)


class StagedEntity(Entity):
    """
    An Entitiy that advances through various stages.

    The entity waits for an external actor to unblock it before it can move onto the next stage.
    """

    definition: StagedEntityDefinition

    def __init__(self, definition: StagedEntityDefinition) -> None:
        """Create a StagedEntity."""
        super(StagedEntity, self).__init__(definition)
        self.current_stage: int = 0
        self.complete: bool = False
        self.stage_blocker: Optional[asyncio.Event] = None
        self.stage_advancement_time_s: float = 0

    def advance_one_stage(self) -> None:
        """Step forward to the next stage."""
        self.current_stage += 1
        if self.current_stage == self.definition.num_stages:
            self.complete = True

    async def _update(self, dt_s: float) -> None:
        """Handles waiting for stage unblocking and waiting for stage completion time."""
        if self.stage_blocker is None:
            # We shouldn't have `_update` called unless `run` is called, which sets up the
            # stage blocker, but added this check for safety.
            return

        # Wait for next stage to be unblocked.
        await self.stage_blocker.wait()

        log.debug(f"{self.definition.name} entered stage {self.current_stage}.")
        await asyncio.sleep(self.definition.stage_time_s)
        log.debug(f"{self.definition.name} ready for next stage.")

        # Clear the event so it can be re-used for the next stage.
        self.stage_blocker.clear()
        self.advance_one_stage()

        if self.complete:
            self.running = False

    async def run(self) -> None:
        """The loop of this Entity that schedules its updates."""
        self.stage_blocker = asyncio.Event()
        await super(StagedEntity, self).run()
        log.debug(f"{self.definition.name} is complete.")

    def __str__(self) -> str:
        """String representation of a StagedEntity."""
        if self.complete:
            text = f"{self.definition.name}: Done"
        else:
            text = f"{self.definition.name}: {self.current_stage}"
        return text



"""Module defining keyboard handlers."""

import logging
from collections import defaultdict
from dataclasses import dataclass, field
from functools import reduce
from typing import Optional, Callable, List, Dict, Iterable, Union

from more_itertools import powerset
from pyglet.event import EVENT_UNHANDLED, EVENT_HANDLED
from pyglet.window.key import (
    symbol_string,
    modifiers_string,
    MOD_NUMLOCK,
    MOD_CAPSLOCK,
    MOD_SCROLLLOCK,
)

import cocos
from .helpers import bitwise_add

NO_MOD = 0

log = logging.getLogger(__name__)


def defaultdict_key_map() -> Dict[int, Dict[Union[int, str], List]]:
    """Return a defaultdict for a key map structure."""
    return defaultdict(lambda: defaultdict(list))


@dataclass(frozen=True)
class ChordDefinition:
    """
    Definition of a key plus a set of modifiers.

    :param key: Scancode of the key. See pyglet.window.key for definitions.
    :param modifiers: Bitwise sum of the modifiers. See pyglet.window.key for definitions.
        Defaults to no modifier.
    :param ignore_modifiers: Tuple of modifiers to ignore. This is useful for ignoring the
        state of common always-on modifiers such as NUMLOCK.
        Defaults to ignoring NUMLOCK, CAPSLOCK and SCROLLLOCK.
    """

    key: int
    modifiers: int = 0
    ignore_modifiers: Iterable[int] = (MOD_NUMLOCK, MOD_CAPSLOCK, MOD_SCROLLLOCK)

    @classmethod
    def build_from_modifier_list(
        cls, key: int, modifiers: List[int]
    ) -> "ChordDefinition":
        """
        Build a Chord from a list of modifiers.

        Combines the modifiers into a single int representing their combined state.
        """
        modifier = reduce(bitwise_add, modifiers)
        return cls(key=key, modifiers=modifier)

    def __str__(self) -> str:
        """
        Return a human readable version of this chord.

        For example:
        ```
        str(ChordDefinition(key=pyglet.window.key.A, modifiers=MOD_CTRL|MOD_SHIFT))
          = "SHIFT+CTRL+A"
        ```
        """
        symbol_str: str = symbol_string(self.key)
        if self.modifiers == 0:
            return symbol_str

        modifier_str: str = modifiers_string(self.modifiers)
        split_modifier_strs: List[str] = modifier_str.split("|")
        stripped_modifier_strs: Iterable[str] = map(
            lambda mod_s: mod_s.replace("MOD_", ""), split_modifier_strs
        )
        return "+".join(stripped_modifier_strs) + "+" + symbol_str


@dataclass(frozen=True)
class KeyboardActionDefinition:
    """
    Definition of callbacks to call when a chord is pressed.

    Optionally takes multiple chords which all trigger the same action.

    A chord can either be a ChordDefinition, which is a specific arrangement of key + modifiers.
    Or it can be a single string character. This is useful when you want to distinguish
    between capital and lowercase letters without handling `MOD_SHIFT` yourself.

    Also note that text events may repeat if the user holds the key down, based on system settings.

    :param chords: List of ChordDefinitions or characters that can trigger the callbacks.
    :param on_select: Called when one of the chords is pressed, or a character is pressed.
    :param on_release: Called when one of the chords is released, or a character is pressed.
    """

    chords: List[Union[ChordDefinition, str]] = field(default_factory=list)

    on_press: Optional[Callable[[], Optional[bool]]] = None
    on_release: Optional[Callable[[], Optional[bool]]] = None


@dataclass
class KeyboardHandlerDefinition:
    """
    Definition of mapping from keyboard inputs to handlers.

    :param key_map: Mapping from chords or characters to actions to take.
        Note that it is probably easier to build the key_map using the methods on this definition,
        such as `add_keyboard_action`, rather than building it yourself.

        The following example shows how the `a` key can have different effects depending on
        the modifier used.

            map = {
                NO_MOD = {
                    A = [KeyboardActionDefinition(on_select=do_one)],
                    ...
                },
                MOD_SHIFT = {
                    A = [KeyboardActionDefinition(on_select=do_two)],
                    ...
                },
                MOD_CTRL | MOD_SHIFT = {
                    A = [KeyboardActionDefinition(on_select=do_three)],
                    ...
                },
                ...
            }
    :param on_text: Called with the string representation of the keyboard press.
    :param on_text_motion: Called with the pyglet representation of text motion.
    :param on_text_motion_select: Called with the pyglet representation of text motion selection.
    :param focus_required: If True, then this handler only responds to keyboard events
        when it has keyboard focus (i.e. `has_keyboard_focus = True`).
    """

    key_map: Dict[int, Dict[Union[int, str], List[KeyboardActionDefinition]]] = field(
        default_factory=defaultdict_key_map
    )
    on_text: Optional[Callable[[str], Optional[bool]]] = None
    on_text_motion: Optional[Callable[[int], Optional[bool]]] = None
    on_text_motion_select: Optional[Callable[[int], Optional[bool]]] = None
    focus_required: bool = True

    def add_keyboard_action(self, keyboard_action: KeyboardActionDefinition) -> None:
        """Add a KeyboardActionDefinition to this keymap."""
        for chord in keyboard_action.chords:
            if isinstance(chord, str):
                modifier_map = self.key_map[NO_MOD]
                if keyboard_action not in modifier_map[chord]:
                    modifier_map[chord].append(keyboard_action)
            else:
                # Add the action definition to every possible combination of the chord modifier
                # and the ignore modifiers. This works because if an ignored modifier is pressed,
                # then the action will be found in the relevant combined modifier dict.
                for mod_set in powerset(chord.ignore_modifiers):
                    modifiers = reduce(bitwise_add, (chord.modifiers, *mod_set))
                    modifier_map = self.key_map[modifiers]
                    if keyboard_action not in modifier_map[chord.key]:
                        modifier_map[chord.key].append(keyboard_action)

    def remove_keyboard_action(self, keyboard_action: KeyboardActionDefinition) -> None:
        """Remove a KeyboardActionDefinition from this keymap."""
        for chord in keyboard_action.chords:
            if isinstance(chord, str):
                modifier_map = self.key_map[NO_MOD]
                if keyboard_action in modifier_map[chord]:
                    modifier_map[chord].remove(keyboard_action)
            else:
                # Remove the action definition from every possible combination of the chord
                # modifier and the ignore modifiers.
                for mod_set in powerset(chord.ignore_modifiers):
                    modifiers = reduce(bitwise_add, (chord.modifiers, *mod_set))
                    modifier_map = self.key_map[modifiers]
                    if keyboard_action in modifier_map[chord.key]:
                        modifier_map[chord.key].remove(keyboard_action)

    def add_keyboard_action_simple(
        self,
        key: Union[int, str],
        action: Callable[[], Optional[bool]],
        modifiers: int = 0,
    ) -> KeyboardActionDefinition:
        """
        Simplified version of `add_keyboard_action`.

        Adds a single callback to occur on key press for the given key.

        :param key: The pyglet keyboard key integer; or text character.
        :param action: The function to call when the key is pressed.
        :param modifiers: The pyglet keyboard modifier integer.
        :return: The `KeyboardActionDefinition` created.
        """
        if isinstance(key, int):
            definition = KeyboardActionDefinition(
                chords=[ChordDefinition(key=key, modifiers=modifiers)], on_press=action,
            )
        else:
            definition = KeyboardActionDefinition(chords=[key], on_press=action,)
        self.add_keyboard_action(definition)
        return definition


class KeyboardHandler(cocos.cocosnode.CocosNode):
    """
    KeyboardHandler is a cocosnode that handles keyboard events.

    The handler has reference to a KeyboardHandlerDefinition which defines what actions to take
    when keys (plus optional modifiers) are pressed or released.

    Multiple KeyboardHandlers can exist at once. They will receive keyboard events
    following the normal cocosnode event handler rules.
    """

    def __init__(self, definition: KeyboardHandlerDefinition):
        """
        Create a new KeyboardHandler.

        :param definition: The definition of key to action mapping.
        """
        super(KeyboardHandler, self).__init__()
        self.has_keyboard_focus: bool = False
        self.definition = definition

    def on_enter(self):
        """Called every time just before the node enters the stage."""
        cocos.director.director.window.push_handlers(self)
        super(KeyboardHandler, self).on_enter()

    def on_exit(self):
        """Called every time just before the node exits the stage."""
        cocos.director.director.window.remove_handlers(self)
        super(KeyboardHandler, self).on_exit()

    @property
    def should_handle_keyboard_event(self) -> bool:
        """Return True if this KeyboardHandler should handle keyboard events, otherwise False."""
        return not self.definition.focus_required or self.has_keyboard_focus

    def set_focused(self):
        """Set this handler to be focused."""
        self.has_keyboard_focus = True

    def set_unfocused(self):
        """Set this handler to not be focused."""
        self.has_keyboard_focus = False

    def on_key_press(self, symbol: int, modifiers: int) -> Optional[bool]:
        """
        Called when the user presses a key.

        :param symbol: Integer representation of the key that was pressed.
            See pyglet.window.key for key code definitions.
        :param modifiers: Bitwise sum of the modifiers that are currently pressed.
            See pyglet.window.key for key code definitions.
        :return: True if the event was handled by any KeyboardAction, otherwise None.
        """
        if not self.should_handle_keyboard_event:
            return EVENT_UNHANDLED

        results = []
        modifier_map = self.definition.key_map[modifiers]
        for handler in modifier_map[symbol]:
            if handler.on_press is not None:
                results.append(handler.on_press())

        if EVENT_HANDLED in results:
            log.debug(f"on_key_press consumed by {self}")
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def on_key_release(self, symbol: int, modifiers: int) -> Optional[bool]:
        """
        Called when the user releases a key.

        :param symbol: Integer representation of the key that was releases.
            See pyglet.window.key for key code definitions.
        :param modifiers: Bitwise sum of the modifiers that are currently released.
            See pyglet.window.key for key code definitions.
        :return: True if the event was handled by any KeyboardAction, otherwise None.
        """
        if not self.should_handle_keyboard_event:
            return EVENT_UNHANDLED

        results = []
        modifier_map = self.definition.key_map[modifiers]
        for handler in modifier_map[symbol]:
            if handler.on_release is not None:
                results.append(handler.on_release())

        if EVENT_HANDLED in results:
            log.debug(f"on_key_release consumed by {self}")
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def on_text(self, text: str) -> Optional[bool]:
        """
        Called with the character that the user pressed.

        If `on_text` is set in the definition, that callback takes precedence over the keymap.

        This is called with the string representation of the key that was pressed,
        accounting for modifiers. For example, lowercase and capital letters are handled.

        Typically this is called after on_key_press and before on_key_release,
        but may also be called multiple times if the key is held down (key repeating);
        or called without key presses if another input method was used (e.g., a pen input).

        :param text: Single character string that was pressed.
        :return: True if the event was handled by any KeyboardAction, otherwise None.
        """
        if not self.should_handle_keyboard_event:
            return EVENT_UNHANDLED

        if self.definition.on_text is not None:
            result = self.definition.on_text(text)
            if result is EVENT_HANDLED:
                log.debug(f"on_text({text!r}) consumed by {self}")
                return EVENT_HANDLED

        results = []
        modifier_map = self.definition.key_map[NO_MOD]
        for handler in modifier_map[text]:
            if handler.on_press is not None:
                results.append(handler.on_press())
            if handler.on_release is not None:
                results.append(handler.on_release())

        if EVENT_HANDLED in results:
            log.debug(f"on_text({text!r}) consumed by {self}")
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def _should_handle_on_text_motion(self, motion: int) -> bool:
        """
        Determine if this Box should attempt to handle the mouse press event.

        :param motion: Int indicating which text motion has occurred.
        :return: True if this node should handle the text motion event.
        """
        return (
            self.should_handle_keyboard_event
            and self.definition.on_text_motion is not None
        )

    def _should_handle_on_text_motion_select(self, motion: int) -> bool:
        """
        Determine if this Box should attempt to handle the mouse press event.

        :param motion: Int indicating which text motion has occurred.
        :return: True if this node should handle the text motion select event.
        """
        return (
            self.should_handle_keyboard_event
            and self.definition.on_text_motion_select is not None
        )

    def on_text_motion(self, motion: int) -> Optional[bool]:
        """
        Called when text motion inputs are pressed.

        This handles typical text motion combination inputs, such as ctrl+home.

        See https://pyglet-current.readthedocs.io/en/latest/programming_guide/keyboard.html#motion-events.  # noqa

        :param motion: The pyglet constant representing the text motion that occurred.
        """
        if not self._should_handle_on_text_motion(motion):
            return EVENT_UNHANDLED

        if self.definition.on_text_motion is not None:
            self.definition.on_text_motion(motion)
            return EVENT_HANDLED
        return EVENT_UNHANDLED

    def on_text_motion_select(self, motion: int) -> Optional[bool]:
        """
        Called when text motion inputs are pressed.

        This handles typical text motion combination inputs, such as ctrl+home.

        See https://pyglet-current.readthedocs.io/en/latest/programming_guide/keyboard.html#motion-events.  # noqa

        :param motion: The pyglet constant representing the text motion that occurred.
        """
        if not self._should_handle_on_text_motion_select(motion):
            return EVENT_UNHANDLED

        if self.definition.on_text_motion_select is not None:
            self.definition.on_text_motion_select(motion)
            return EVENT_HANDLED
        return EVENT_UNHANDLED


def add_simple_keyboard_handler(
    parent: cocos.cocosnode.CocosNode,
    key: Union[int, str],
    action: Callable[[], Optional[bool]],
) -> KeyboardHandler:
    """
    Add a simple keyboard handler to the given parent.

    This allows the node to trigger the given action when the given key is pressed, regardless
    of which keyboard modifiers are currently pressed.

    This requires that the parent has been made focusable because this is intended to add a
    node-specific keyboard action, rather than a global action. This can be done by either adding
    a FocusBox to the parent or using the simpler `make_focusable` method.
    """
    keyboard_handler_definition = KeyboardHandlerDefinition(focus_required=True)
    keyboard_handler_definition.add_keyboard_action_simple(key, action)
    keyboard_handler = KeyboardHandler(keyboard_handler_definition)
    parent.add(keyboard_handler)
    return keyboard_handler



"""An Entity that does work on another entity."""

import asyncio
import logging

from shimmer.engine.entity.definition import EntityDefinition
from shimmer.engine.entity.entity import Entity
from shimmer.engine.staged_entity.entity import StagedEntity

log = logging.getLogger(__name__)


class Worker(Entity):
    """An Entity that does work on another entity."""

    def __init__(self, target: StagedEntity) -> None:
        """
        Create a Worker.

        :param target: The target entity to work on.
        """
        super(Worker, self).__init__(EntityDefinition(name="worker"))
        self.target = target
        self.rest_period = 0

    async def _update(self, dt_s: float) -> None:
        """Action to perform on every update tick."""
        if self.target.stage_blocker and not self.target.stage_blocker.is_set():
            log.debug(f"Starting work on {self.target.definition.name}")
            await asyncio.sleep(self.target.stage_advancement_time_s)
            self.target.stage_blocker.set()
            log.debug(f"Finished working on {self.target.definition.name}")
            await asyncio.sleep(self.rest_period)
        else:
            log.debug(f"Worker {id(self)!r} has no work to do.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shimmer',
 'shimmer.display',
 'shimmer.display.components',
 'shimmer.display.programmable',
 'shimmer.display.widgets',
 'shimmer.engine',
 'shimmer.engine.entity',
 'shimmer.engine.programmable',
 'shimmer.engine.staged_entity',
 'shimmer.engine.widgets']

package_data = \
{'': ['*']}

install_requires = \
['cocos2d>=0.6.7,<0.7.0',
 'janus>=0.4.0,<0.5.0',
 'more-itertools>=8.0.2,<9.0.0',
 'pyglet==1.4.3',
 'toml>=0.9,<0.10']

setup_kwargs = {
    'name': 'shimmer',
    'version': '1.3.0',
    'description': 'Create games in python, without hassle - batteries included!',
    'long_description': 'Shimmer\n-------\n\n![badge](https://github.com/MartinHowarth/shimmer/workflows/Test/badge.svg)\n\nHello!\n\nTesting\n-------\nThe following command should be used to run the tests:\n\n    poetry run pytest tests\n\nFor tests where a window is displayed, read the test description and\npress `Y` or `N` to pass or fail the test. This is intended to allow humans\nto validate that components look and behave as expected.\n\nMost components have non-graphical testing as well that covers the event handling, \nbut that is no replacement for a real human deciding whether it looks and feels good!\n\n## Running all non-graphical tests.\nSet `SKIP_GUI_TESTS=1` in your environment to skip graphical tests.\n\nThis skips all tests that require a GUI and user interaction.\n',
    'author': 'Martin Howarth',
    'author_email': 'howarth.martin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MartinHowarth/shimmer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Module defining a close button that removes its parent Node from the scene when clicked."""

from dataclasses import replace

from shimmer.display.data_structures import Color
from shimmer.display.helpers import bundle_callables
from shimmer.display.widgets.button import ButtonDefinition, Button

CloseButtonDefinitionBase = ButtonDefinition(
    text="X", base_color=Color(200, 0, 127), hover_color=Color(255, 0, 127)
)


class CloseButton(Button):
    """Standard close button that removes its parent CocosNode when clicked."""

    def __init__(self, definition: ButtonDefinition):
        """Create a CloseButton."""
        if definition.on_release is not None:
            on_release = bundle_callables(definition.on_release, self._kill_parent)
        else:
            on_release = self._kill_parent
        defn = replace(definition, on_release=on_release)
        super(CloseButton, self).__init__(defn)

    def _kill_parent(self, *_, **__):
        """Kill the parent of this button, which will kill all its children too."""
        self.parent.kill()



"""Graphical display of a set of programmable code instructions."""

import logging
from dataclasses import dataclass, field
from typing import List, cast

from shimmer.display.components.box import BoxDefinition
from shimmer.display.components.box_layout import BoxColumn
from shimmer.display.programmable.instruction import (
    InstructionDisplay,
    InstructionDisplayDefinition,
)
from shimmer.engine.programmable.definition import CodeBlock

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class CodeBlockDisplayDefinition(BoxDefinition):
    """Definition of how to display a CodeBlock."""

    instruction_definition: InstructionDisplayDefinition = field(
        default_factory=InstructionDisplayDefinition
    )
    # Spacing between the vertically arranged instructions.
    spacing: int = 0


class CodeBlockDisplay(BoxColumn):
    """Graphical display of a set of programmable code instructions."""

    def __init__(
        self, code_block: CodeBlock, code_block_definition: CodeBlockDisplayDefinition,
    ):
        """
        Create a new CodeBlockDisplay.

        :param code_block: The code block to display.
        """
        super(CodeBlockDisplay, self).__init__([], code_block_definition.spacing)
        self.code_block: CodeBlock = code_block
        self.definition: CodeBlockDisplayDefinition = code_block_definition

        self._boxes = cast(List[InstructionDisplay], self._boxes)
        self.update_instructions()

    def update_instructions(self):
        """Re-create all instruction displays."""
        # We could enhance this later to only re-create those that have
        # changed, but keep it simple for starters.
        for instruction_display in self._boxes:
            self.remove(instruction_display)

        for instruction in self.code_block.instructions:
            new_display = InstructionDisplay(
                instruction, self.definition.instruction_definition
            )
            # Insert at start of list so we get a top-to-bottom list, which matches how the code
            # will be executed.
            self.add(new_display, position=0)



"""Module defining windows."""

from dataclasses import dataclass, replace
from typing import Dict, Optional, Union, Tuple

import cocos
from ..alignment import (
    PositionalAnchor,
    CenterCenter,
    CenterTop,
)
from ..components.box import Box, BoxDefinition, ActiveBox
from ..components.draggable_box import DraggableBox, DraggableBoxDefinition
from ..components.focus import make_focusable
from ..components.font import FontDefinition, Calibri
from ..components.mouse_box import (
    MouseClickEventCallable,
    MouseBox,
    MouseVoidBoxDefinition,
)
from ..data_structures import Color
from ..primitives import create_color_rect
from ..widgets.button import Button
from ..widgets.close_button import (
    CloseButton,
    CloseButtonDefinitionBase,
)
from ..widgets.text_box import TextBox, TextBoxDefinition


@dataclass(frozen=True)
class WindowDefinition(BoxDefinition):
    """
    Definition of the visual style of a Window.

    Inherits from MouseVoidBoxDefinition so that the default window swallows all mouse events
    that happen within it. Other components can be added into the window that can receive
    those events first. This prevents users clicking on a window and something underneath
    the window receiving the event accidentally.

    The `height` parameter is set on init to be the sum of the `body_height` and the
    `title_bar_height`.
    """

    width: int = 200

    # Height of the window body, excluding the title bar.
    body_height: int = 200

    title: Optional[str] = None
    title_font_definition: FontDefinition = Calibri

    # Height of the title bar. If None then it will match the height of the title font.
    title_bar_height: Optional[int] = None

    # Space to leave between buttons in the title bar
    title_bar_button_spacing: int = 2
    title_bar_color: Color = Color(20, 120, 255)
    background_color: Color = Color(20, 20, 20)

    # Callback to call when the window is closed using the close button.
    on_close: Optional[MouseClickEventCallable] = None

    def get_title_bar_height(self) -> int:
        """
        Height of the title bar.

        This is the defined height `title_bar_height` if given,
        otherwise the height of the title font.
        """
        if self.title_bar_height is not None:
            return self.title_bar_height
        return self.title_font_definition.height

    def __post_init__(self):
        """
        Initialise the `height` of the window.

        This sets height to the sum of the body and title bar heights.
        """
        # Have to use this __setattr__ workaround because the dataclass is frozen.
        object.__setattr__(
            self, "height", self.body_height + self.get_title_bar_height()
        )


class Window(ActiveBox):
    """
    A Window is a draggable, closeable Box with a title bar.

    It has an inner_box for containing children for displaying inside the window.

    All nodes added to the window are placed above this VoidBox so they receive the mouse events
    first.
    """

    def __init__(self, definition: WindowDefinition):
        """
        Create a Window.

        :param definition: Definition of the Window style.
        """
        self.definition: WindowDefinition = definition
        self.title_bar_height = self.definition.get_title_bar_height()

        super(Window, self).__init__(definition)

        self._title: Optional[cocos.text.Label] = None
        self._title_boxes: Dict[str, Box] = {}
        self._title_bar_background: Optional[cocos.layer.ColorLayer] = None
        self._update_title()
        self._update_title_bar_background()
        self.update_background()
        self._update_close_button()
        self._update_drag_zone()
        self.focus_box = make_focusable(self)

        # Add the inner box, which is the main body of the window excluding the title bar.
        self.body: Box = Box(
            BoxDefinition(
                width=self.definition.width, height=self.definition.body_height,
            )
        )
        self.add(self.body)
        self.add(
            MouseBox(
                MouseVoidBoxDefinition(
                    width=self.definition.width, height=self.definition.height,
                )
            ),
            z=-10000,
        )

    @property
    def _title_bar_button_height(self):
        """Height of the title bar buttons, accounting for spacing."""
        return self.title_bar_height - (2 * self.definition.title_bar_button_spacing)

    @property
    def _leftmost_title_bar_button_position(self) -> int:
        """Return the leftmost edge of the buttons in the title bar."""
        return min(
            [
                self.rect.width - box.x
                for box in self._title_boxes.values()
                if isinstance(box, Button)
            ]
        )

    def _create_square_title_button_rect(self, x: int) -> cocos.rect.Rect:
        """
        Rect defining a square title bar button.

        :param x: X position of the button.
        :return: Rect defining a square title bar button.
        """
        edge_length = self._title_bar_button_height
        return cocos.rect.Rect(
            x,
            self.rect.height
            - self.title_bar_height
            + self.definition.title_bar_button_spacing,
            edge_length,
            edge_length,
        )

    def _update_title(self):
        """Recreate the title."""
        if self._title is not None:
            self.remove(self._title)

        if self.definition.title is None:
            return

        title_definition = TextBoxDefinition(
            text=self.definition.title,
            height=self.definition.title_bar_height,
            width=self.definition.width,
        )

        self._title = TextBox(title_definition)
        self.add(self._title)
        self._title.align_anchor_with_other_anchor(self, CenterTop)
        self._title.x = 10

    def _update_title_bar_background(self):
        """Redefine the background color of the title bar."""
        if self._title_bar_background is not None:
            self.remove(self._title_bar_background)

        self._title_bar_background = create_color_rect(
            self.rect.width, self.title_bar_height, self.definition.title_bar_color,
        )
        self._title_bar_background.position = (
            0,
            self.rect.height - self.title_bar_height,
        )
        self.add(self._title_bar_background, z=-1)

    def _update_close_button(self):
        """Redefine the window close button."""
        close_button_rect = self._create_square_title_button_rect(
            self.rect.width
            - (self._title_bar_button_height + self.definition.title_bar_button_spacing)
        )
        definition = replace(
            CloseButtonDefinitionBase,
            width=close_button_rect.width,
            height=close_button_rect.height,
            on_release=self.definition.on_close,
        )
        close_button = CloseButton(definition)
        close_button.position = close_button_rect.position
        self._update_title_bar_box("close", close_button)

    @property
    def close_button(self) -> CloseButton:
        """Get the close button of this window."""
        return self._title_boxes["close"]

    def _update_drag_zone(self):
        """
        Redefine the draggable area of the title bar.

        This creates a draggable area of the window covering the entire title bar to the left of
        the leftmost title bar button.
        """
        drag_box_definition = DraggableBoxDefinition(
            width=self.rect.width - self._leftmost_title_bar_button_position,
            height=self.title_bar_height,
        )
        self._update_title_bar_box("drag", DraggableBox(drag_box_definition))
        self._title_boxes["drag"].position = 0, self.rect.height - self.title_bar_height

    def _update_title_bar_box(self, name: str, box: Box) -> None:
        """
        Create or re-create a title bar Box known by the given name.

        :param name: Internal identifier of the Box.
        :param box: New Box node to place in the title bar.
        """
        if name in self._title_boxes.keys():
            self.remove(self._title_boxes[name])

        self.add(box)
        self._title_boxes[name] = box

    def add_child_to_body(
        self,
        child: Box,
        body_anchor: PositionalAnchor = CenterCenter,
        child_anchor: Optional[PositionalAnchor] = None,
        spacing: Union[int, Tuple[int, int], cocos.draw.Point2] = 0,
    ) -> None:
        """
        Add a child to the body of the window aligned to the edges or center of the window.

        See Box.align_anchor_with_other_anchor for fuller documentation.

        If you want finer control over spacing, then you should set positions and add children
        directly to the `body`.

        :param child: The child Box to add.
        :param body_anchor: The anchor point of the window body to align to.
            Defaults to CenterCenter.
        :param child_anchor: The anchor point of the child to align.
            Defaults to match `body_anchor`.
        :param spacing: Pixels to leave between the anchors.
        """
        child.align_anchor_with_other_anchor(
            self.body, body_anchor, child_anchor, spacing
        )
        self.body.add(child)

    def close(self):
        """Close this window by simulating a click on the close button."""
        self.close_button.on_keyboard_select_release()



"""Visual display of a code block If/Else/Elif statement."""

import logging
from dataclasses import dataclass, field
from typing import cast, Optional, List

from shimmer.display.components.box import Box, BoxDefinition
from shimmer.display.components.box_layout import BoxColumn
from shimmer.display.programmable.code_block import (
    CodeBlockDisplay,
    CodeBlockDisplayDefinition,
)
from shimmer.display.programmable.instruction import (
    InstructionDisplay,
    InstructionDisplayDefinition,
)
from shimmer.engine.programmable.definition import (
    InstructionWithCodeBlock,
    IfElifElse,
)

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class InstructionWithBlockDisplayDefinition(BoxDefinition):
    """Definition of how to display an instruction with an associated CodeBlock."""

    # Definition of the main If statement display
    instruction_definition: InstructionDisplayDefinition = field(
        default_factory=InstructionDisplayDefinition
    )
    # Definition of the code block display within the If statement.
    code_block_definition: CodeBlockDisplayDefinition = field(
        default_factory=CodeBlockDisplayDefinition
    )
    # How far the code block is indented by compared to the If instruction.
    code_block_indentation: int = 20


class InstructionWithBlockDisplay(Box):
    """
    Graphical display of an instruction with a nested code block.

    Consists of:
      - an InstructionDisplay
      - a CodeBlockDisplay below that, indented by a configurable amount.
    """

    def __init__(
        self,
        instruction: InstructionWithCodeBlock,
        definition: InstructionWithBlockDisplayDefinition,
    ):
        """
        Create an IfElifElseDisplay.

        :param instruction: The InstructionWithCodeBlock instruction to display.
        :param definition: The definition of the display style to use for the code block.
        """
        super(InstructionWithBlockDisplay, self).__init__()
        self.instruction: InstructionWithCodeBlock = instruction
        self.definition: InstructionWithBlockDisplayDefinition = definition
        self.instruction_display: Optional[InstructionDisplay] = None
        self.code_block_display: Optional[CodeBlockDisplay] = None
        self.update_boxes()

    def update_boxes(self):
        """Recreate the child boxes of this display."""
        if self.instruction_display is not None:
            self.remove(self.instruction_display)
        if self.code_block_display is not None:
            self.remove(self.code_block_display)
        self.code_block_display = CodeBlockDisplay(
            self.instruction.code_block, self.definition.code_block_definition
        )
        self.code_block_display.position = self.definition.code_block_indentation, 0
        self.instruction_display = InstructionDisplay(
            self.instruction, self.definition.instruction_definition
        )
        self.instruction_display.position = 0, self.code_block_display.rect.height
        self.add(self.instruction_display)
        self.add(self.code_block_display)


class IfElifElseDisplay(BoxColumn):
    """
    A vertical arrangement of InstructionWithBlockDisplays to display an If/Elif/Else block.

    Produces output similar to:
    if method():
        if_method1()
    elif method2():
        elif_method()
    else:
        pass
    """

    def __init__(
        self,
        instruction: IfElifElse,
        definition: InstructionWithBlockDisplayDefinition,
    ):
        """
        Create an IfElifElseDisplay.

        :param instruction: The IfElifElse instruction to display.
        :param definition: The definition of the display style to use for each code block.
        """
        super(IfElifElseDisplay, self).__init__([], 0)
        self.instruction: IfElifElse = instruction
        self.definition: InstructionWithBlockDisplayDefinition = definition

        self._boxes = cast(List[InstructionWithBlockDisplay], self._boxes)

        self.update_boxes()

    def update_boxes(self):
        """Recreate the child boxes of this display."""
        for display in self._boxes:
            self.remove(display)

        for _elif in self.instruction.elifs:
            elif_display = InstructionWithBlockDisplay(_elif, self.definition)
            # Insert at start of list so we get a top-to-bottom list, which matches how the code
            # will be executed.
            self.add(elif_display, position=0)

        if self.instruction.else_ is not None:
            else_display = InstructionWithBlockDisplay(
                self.instruction.else_, self.definition
            )
            # Insert at start of list so we get Else at the bottom.
            self.add(else_display, position=0)

        # Add instruction last so it appears at the top.
        instruction_display = InstructionWithBlockDisplay(
            self.instruction, self.definition
        )
        self.add(instruction_display)



"""Definition of a Box that tracks the cursor."""

from dataclasses import dataclass
from typing import Optional

import cocos
from shimmer.display.components.box import ActiveBox, BoxDefinition
from shimmer.display.primitives import Point2d


@dataclass(frozen=True)
class BoxTrackCursorDefinition(BoxDefinition):
    """
    Definition of a Box that tracks the cursor.

    :param offset: (x, y) offset from the cursor position to set the Box position to.
    """

    offset: Point2d = (0, 0)


class BoxTrackCursor(ActiveBox):
    """A Box that tracks the cursor."""

    definition_type = BoxTrackCursorDefinition

    def __init__(self, definition: Optional[BoxTrackCursorDefinition] = None):
        """Create a BoxTrackCursor."""
        super(BoxTrackCursor, self).__init__(definition)
        self.definition: BoxTrackCursorDefinition = self.definition

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        """
        Cocos director callback when the mouse is moved.

        Updates the position of this Box to the cursor position.

        Does not capture the event, so it may be handled by other entities as well.
        """
        coord = cocos.director.director.get_virtual_coordinates(x, y)
        # Position is relative to the parent, so need the mouse coordinates translated into the
        # local space of the parent to determine the correct relative position of this Box.
        self.position = self.parent.point_to_local(coord) + self.definition.offset



"""Definition of a StagedEntity."""

from dataclasses import dataclass

from ..entity.definition import EntityDefinition


@dataclass
class StagedEntityDefinition(EntityDefinition):
    """Definition of a StagedEntity."""

    num_stages: int
    total_time_s: int

    @property
    def stage_time_s(self):
        """Seconds taken for each stage to complete, assuming all stages are of equal length."""
        return self.total_time_s / self.num_stages
